"""
Helpers to load checkpoints for learned feature maps (attentions) or other parameters
"""
import torch
import torch.nn as nn
from omegaconf import OmegaConf

from src.utils.logging import print_header, _format_arg
from .convert_model import convert_attention
from .peft import create_peft_config


def load_and_convert_attns(model: nn.Module,
                           model_config: dict,
                           attention_type: str = None,
                           checkpoint_path: str = None,
                           print_model: bool = False,
                           merge_loras: bool = False,
                           train_converted: bool = True,  # Should be false if loading distill checkpoint by default
                           peft_gradient_checkpointing: bool = None,
                           train_attention: bool = False,  # Should be true if converting attentions for first time,
                           freeze_weights: bool = True,
                           rank: int = 0,
                           remove_base_attn: bool = True,
                          ) -> nn.Module:
    """
    Load trained attention kernel parameter weights
    """
    if freeze_weights:
        for p in model.parameters():
            p.requires_grad = False

    if attention_type is not None:  # override default
        model_config['attention']['attention_type'] = attention_type
    model_config['attention']['rank'] = rank   # multi-gpu debugging

    model = convert_attention(model, model_config['attention'], 
                              train_attention, remove_base_attn)

    # Add low-rank adapters
    peft_key = 'peft'  # inconsistency across configs... why do this to myself
    if 'peft_config' in model_config['attention']:
        peft_key = 'peft_config'
    if peft_key in model_config['attention']:
        peft_config = model_config['attention'][peft_key]
        model, peft_config = create_peft_config(model, peft_config, 
                                                model_config['model']['torch_dtype'],
                                                preserve_requires_grad=train_converted,
                                                use_gradient_checkpointing=peft_gradient_checkpointing)
    else:
        peft_config = None

    if print_model and rank == 0:  # Look at model
        print_header('*** Model before checkpoint load ***')
        print(model)

    # Load any trained attentions
    if checkpoint_path is not None:
        print(f'Loading weights from {checkpoint_path}...')
        state_dict = torch.load(checkpoint_path)['model_state_dict']
        _keys = model.load_state_dict(state_dict, strict=False)
        try:
            assert len(_keys.unexpected_keys) == 0
            if rank == 0:
                print_header('*** All expected keys matched successfully ***')
                if print_model:
                    for k in state_dict.keys():
                        print(k)
        except Exception as e:
            if rank == 0:
                print(e)
                print_header('*** Error: unexpected keys in checkpoint ***')
                print('Unexpected keys:')
                for k in _keys.unexpected_keys:
                    print(k)
    if print_model and rank == 0:  # Look at model
        print_header('*** Model ***')
        print(model)
    if merge_loras:
        model = model.merge_and_unload()
        if print_model and rank == 0:
            print_header('*** Model (after merging adapters) ***')
            print(model)
    if print_model and rank == 0:  # Look at model
        print_header('*** Trainable Parameters ***')
        for n, p in model.named_parameters():
            if p.requires_grad:
                print(f'├── {n} (dtype = {p.dtype})')
    return model, peft_config


def load_and_convert_finetune(model: nn.Module,
                              finetune_config: dict,
                              checkpoint_path: str = None,
                              print_model: bool = False,
                              merge_loras: bool = False,
                              peft_gradient_checkpointing: bool = None,
                              rank: int = 0,
                              **peft_kwargs: any):
    """
    Load trained adapter / model weights
    """
    # Add low-rank adapters
    peft_config = None
    if finetune_config.finetune.method == 'lora':
        if getattr(finetune_config.finetune, 'kwargs', None) is not None:
            model, peft_config = create_peft_config(
                model, finetune_config.finetune,
                use_gradient_checkpointing=peft_gradient_checkpointing,
                **peft_kwargs,
            )
        # Keep specified weights trainable
        if 'trainable_weights' in finetune_config.finetune:
            for name in finetune_config.finetune['trainable_weights']:
                for n, p in model.named_parameters():
                    if name in n:
                        p.requires_grad = True
    else:
        for p in model.parameters():
            p.requires_grad = False
        # Keep specified weights trainable
        if 'trainable_weights' in finetune_config.finetune:
            for name in finetune_config.finetune['trainable_weights']:
                for n, p in model.named_parameters():
                    if name in n: 
                        if 'layers_to_ignore' in finetune_config.finetune:
                            layer = int(n.split('layers.')[-1].split('.')[0])
                            if layer not in finetune_config.finetune['layers_to_ignore']:
                                p.requires_grad = True
                        else:
                            p.requires_grad = True
        

    # Load weights
    if checkpoint_path:
        state_dict = torch.load(checkpoint_path)['model_state_dict']
        _keys = model.load_state_dict(state_dict, strict=False)
        try:
            assert len(_keys.unexpected_keys) == 0
            if rank == 0:
                print_header('*** All expected keys matched successfully ***')
        except Exception as e:
            if rank == 0:
                print(e)
                print_header('*** Error: unexpected keys in checkpoint ***')
                print('Unexpected keys:')
                for k in _keys.unexpected_keys:
                    print(k)

    if print_model and rank == 0:  # Look at model
        print_header('*** Model ***')
        print(model)

    if merge_loras:
        try:
            model = model.merge_and_unload()
            if print_model and rank == 0:
                print_header('*** Model (after merging adapters) ***')
                print(model)
        except Exception as e:
            print(e)

    if print_model and rank == 0:  # Look at model
        print_header('*** Trainable Parameters ***')
        count = 0
        for n, p in model.named_parameters():
            if p.requires_grad:
                print(f'├── {n}.requires_grad: {p.requires_grad}')
                count += 1
        if count == 0:
            print('(none)')

    return model, peft_config



import torch
import math
import sys

try:
    import thunderkittens as tk
except:
    print("Could not import thunderkittens")


def ref_fftconv(u, k, N):
    L = u.shape[-1]
    u_f = torch.fft.fft(u.float(), n = N)
    k_f = torch.fft.fft(k.float(), n = N)
    y_f = u_f * k_f
    y = torch.fft.ifft(y_f, n = N).real[..., :L].to(u.dtype).contiguous()
    return y

def fft_matrix(N):
    n = torch.arange(N)
    k = n.view(-1, 1)
    M = torch.exp(-2j * torch.pi * n * k / N)
    return M

def compute_twiddle_factors_fft(n, m):
    """Compute the twiddle factors of size n x m"""
    n_a = torch.arange(n).view(-1, 1)
    m_a = torch.arange(m)
    N = n * m
    M = torch.exp(-2j * torch.pi * n_a * m_a / N)
    return M

def ifft_matrix(N):
    n = torch.arange(N)
    k = n.view(-1, 1)
    M = torch.exp(2j * torch.pi * n * k / N)
    return M

def compute_twiddle_factors_ifft(n, m):
    """Compute the twiddle factors of size n x m"""
    n_a = torch.arange(n).view(-1, 1)
    m_a = torch.arange(m)
    N = n * m
    M = torch.exp(2j * torch.pi * n_a * m_a / N)
    return M

def generate_1024(
    f_real, f_imag,
    finv_real, finv_imag,
    tw_real, tw_imag,
    twinv_real, twinv_imag,
    kfT_real, kfT_imag,
    H, N1
):
    # fft matrix
    f_sqrt_64x64_real = torch.zeros(64, 64, device=f_real.device).to(f_real.dtype)
    f_sqrt_64x64_real[:N1, :N1] = f_real
    f_sqrt_64x64_real[N1:, N1:] = f_real

    f_sqrt_64x64_imag = torch.zeros(64, 64, device=f_imag.device).to(f_imag.dtype)
    f_sqrt_64x64_imag[N1:, N1:] = f_imag
    f_sqrt_64x64_imag[:N1, :N1] = f_imag

    # ifft matrix
    f_inv_sqrt_64x64_real = torch.zeros(64, 64, device=finv_real.device).to(finv_real.dtype)
    f_inv_sqrt_64x64_real[:N1, :N1] = finv_real
    f_inv_sqrt_64x64_real[N1:, N1:] = finv_real

    f_inv_sqrt_64x64_imag = torch.zeros(64, 64, device=finv_imag.device).to(finv_imag.dtype)
    f_inv_sqrt_64x64_imag[N1:, N1:] = finv_imag
    f_inv_sqrt_64x64_imag[:N1, :N1] = finv_imag

    # Normalization factor to make IFFT exact inverse of FFT
    tw_real_64x64_real = torch.zeros(64, 64, device=tw_real.device).to(tw_real.dtype)
    tw_real_64x64_real[:N1, :N1] = tw_real
    tw_real_64x64_real[N1:, N1:] = tw_real
    tw_real_64x64_real[:N1, N1:] = tw_real
    tw_real_64x64_real[N1:, :N1] = tw_real

    tw_imag_64x64_imag = torch.zeros(64, 64, device=tw_imag.device).to(tw_imag.dtype)
    tw_imag_64x64_imag[:N1, :N1] = tw_imag
    tw_imag_64x64_imag[N1:, N1:] = tw_imag
    tw_imag_64x64_imag[:N1, N1:] = tw_imag
    tw_imag_64x64_imag[N1:, :N1] = tw_imag

    # twiddle inverse
    twinv_real_64x64_real = torch.zeros(64, 64, device=twinv_real.device).to(twinv_real.dtype)
    twinv_real_64x64_real[:N1, :N1] = twinv_real
    twinv_real_64x64_real[N1:, N1:] = twinv_real
    twinv_real_64x64_real[:N1, N1:] = twinv_real
    twinv_real_64x64_real[N1:, :N1] = twinv_real

    twinv_imag_64x64_imag = torch.zeros(64, 64, device=twinv_imag.device).to(twinv_imag.dtype)
    twinv_imag_64x64_imag[:N1, :N1] = twinv_imag
    twinv_imag_64x64_imag[N1:, N1:] = twinv_imag
    twinv_imag_64x64_imag[:N1, N1:] = twinv_imag
    twinv_imag_64x64_imag[N1:, :N1] = twinv_imag

    # Compute the regular FFT if the seq len isn't 512 or 2048
    k_fT_64x64_real = torch.zeros(H, 64, 64, device=kfT_real.device).to(torch.bfloat16)
    k_fT_64x64_real[:, :N1, :N1] = kfT_real.to(k_fT_64x64_real.dtype)
    k_fT_64x64_real[:, N1:, N1:] = kfT_real.to(k_fT_64x64_real.dtype)
    k_fT_64x64_real[:, :N1, N1:] = kfT_real.to(k_fT_64x64_real.dtype)
    k_fT_64x64_real[:, N1:, :N1] = kfT_real.to(k_fT_64x64_real.dtype)

    k_fT_64x64_imag = torch.zeros(H, 64, 64, device=kfT_imag.device).to(torch.bfloat16)
    k_fT_64x64_imag[:, :N1, :N1] = kfT_imag.to(k_fT_64x64_imag.dtype)
    k_fT_64x64_imag[:, N1:, N1:] = kfT_imag.to(k_fT_64x64_imag.dtype)
    k_fT_64x64_imag[:, :N1, N1:] = kfT_imag.to(k_fT_64x64_imag.dtype)
    k_fT_64x64_imag[:, N1:, :N1] = kfT_imag.to(k_fT_64x64_imag.dtype)
    
    return f_sqrt_64x64_real, f_sqrt_64x64_imag, f_inv_sqrt_64x64_real, f_inv_sqrt_64x64_imag, tw_real_64x64_real, tw_imag_64x64_imag, twinv_real_64x64_real, twinv_imag_64x64_imag, k_fT_64x64_real, k_fT_64x64_imag


class TKFFTConv(torch.nn.Module):
    def __init__(self, k, seqlen=1024, H=128, dtype=torch.bfloat16, use_h100=True):
        super().__init__()
        assert dtype == torch.bfloat16

        self.N = seqlen
        self.H = H
        self.N1 = int(math.sqrt(seqlen))
        self.dtype = dtype
        
        f_mat = fft_matrix(self.N1)
        f_real = f_mat.real.to(dtype).contiguous()
        f_imag = f_mat.imag.to(dtype).contiguous()
        self.register_buffer('f_real', f_real)
        self.register_buffer('f_imag', f_imag)

        tw = compute_twiddle_factors_fft(self.N1, self.N1) / self.N
        tw_real = tw.real.to(dtype).contiguous()
        tw_imag = tw.imag.to(dtype).contiguous()
        self.register_buffer('tw_real', tw_real)
        self.register_buffer('tw_imag', tw_imag)

        finv_mat = ifft_matrix(self.N1)
        finv_real = finv_mat.real.to(dtype).contiguous()
        finv_imag = finv_mat.imag.to(dtype).contiguous()
        self.register_buffer('finv_real', finv_real)
        self.register_buffer('finv_imag', finv_imag)

        twinv = compute_twiddle_factors_ifft(self.N1, self.N1)
        twinv_real = twinv.real.to(dtype).contiguous()
        twinv_imag = twinv.imag.to(dtype).contiguous()
        self.register_buffer('twinv_real', twinv_real)
        self.register_buffer('twinv_imag', twinv_imag)

        k_f = torch.fft.fft(k.float(), n=self.N)
        k_fT = k_f.reshape(H, self.N1, self.N1).transpose(-1, -2).contiguous()
        kfT_real = k_fT.real.to(self.dtype).contiguous()
        kfT_imag = k_fT.imag.to(self.dtype).contiguous()
        self.register_buffer('kfT_real', kfT_real)
        self.register_buffer('kfT_imag', kfT_imag)

        if self.N == 1024:
            (
                self.f_real, self.f_imag, 
                self.finv_real, self.finv_imag, 
                self.tw_real, self.tw_imag, 
                self.twinv_real, self.twinv_imag, 
                self.kfT_real, self.kfT_imag
            ) = generate_1024(
                f_real, f_imag,
                finv_real, finv_imag,
                tw_real, tw_imag,
                twinv_real, twinv_imag,
                self.kfT_real, self.kfT_imag,
                H, self.N1
            )
    
    def forward(self, u, k):
        B, H, L = u.shape
        assert L == self.N
        u_real = u.real.reshape(B, H, self.N1, self.N1).to(self.dtype).contiguous()

        out = tk.fftconv(
            u_real, 
            self.kfT_real, self.kfT_imag, 
            self.f_real, self.f_imag, 
            self.finv_real, self.finv_imag,
            self.tw_real, self.tw_imag, 
            self.twinv_real, self.twinv_imag,
            B, self.H, self.N, self.N1
        )
        
        return out.reshape(B, self.H, self.N).contiguous()

        



import torch 
import torch.nn as nn
# from torchvision.ops import StochasticDepth

import thunderkittens as tk

def run_torch(x, residual, drop_path, dropout, norm, residual_in_fp32=False):
    dropped = dropout(x) #drop_path(dropout(x))
    residual = (residual + dropped ) if residual is not None else dropped
    out = norm(residual.to(dtype=norm.weight.dtype))
    residual = residual.to(torch.float32)
    return out, residual 


def run_tk(x, residual, drop_path, dropout, norm, residual_in_fp32=False):
    x = x.to(dtype=torch.bfloat16)
    residual = residual.to(dtype=torch.bfloat16)
    norm_weight = norm.weight.to(dtype=torch.bfloat16)
    norm_bias = norm.bias.to(dtype=torch.bfloat16)

    out, out_resid = tk.fused_layernorm(
        x, residual, 
        norm_weight, norm_bias, 
        float(dropout.p)
    )

    return out, out_resid 


def run_flash(x, residual, drop_path, dropout, norm, residual_in_fp32=True):
    from baselines.layer_norm_triton import layer_norm_fn, RMSNorm

    rowscale = torch.ones(x.shape[:-1], device=x.device, dtype=x.dtype, )
    # drop_path()
    out, residual = layer_norm_fn(
        x,
        norm.weight,
        norm.bias,
        residual=residual,
        eps=norm.eps,
        dropout_p=dropout.p,
        rowscale=rowscale,
        prenorm=True,
        residual_in_fp32=residual_in_fp32,
        is_rms_norm=False
    )
    return out, residual


def run_naive(x, residual, drop_path, dropout, norm, residual_in_fp32=False):
    use_dropout = False
    
    # 1. dropout on x
    if use_dropout: 
        mask = torch.bernoulli(torch.full_like(x, 1 - dropout.p))
        dropped = x * mask / (1 - dropout.p) 
    else:
        dropped = x
    num_zeros = (dropped == 0).sum().item()

    # 2. drop_path 
    pass # ignore for now

    # 3. residual = dropped + residual
    residual_new = ( dropped + residual ) if residual is not None else dropped 

    # 4. norm
    mean = residual_new.mean(dim=-1, keepdim=True)
    var  = residual_new.var(dim=-1, keepdim=True, unbiased=False)
    residual_norm = (residual_new - mean) / torch.sqrt(var + norm.eps)
    x_norm = norm.weight * residual_norm + norm.bias 

    # compare
    if use_dropout:
        dropped_ref = dropout(x)
    else:
        dropped_ref = x
    num_zeros_ref = (dropped_ref == 0).sum().item()
    residual_ref = ( dropped_ref + residual ) if residual is not None else dropped_ref
    x_norm_ref = norm(residual_ref)

    diff = torch.norm(x_norm - x_norm_ref).item()
    print(x_norm[0,0,:16])
    print(x_norm_ref[0,0,:16])
    print(f"Diff custom: {diff}")
    
    print(f"Num zeros: {num_zeros_ref=}, {num_zeros=}")


if __name__ == "__main__":

    b, n, d = 16, 32, 1024
    p = 0.0
    p_path = 0.00

    torch.manual_seed(0)
    x = torch.randn((b, n, d), device='cuda')
    residual = torch.randn((b, n, d), device='cuda')

    # manual impl.
    norm = nn.LayerNorm(d).cuda()
    dropout = nn.Dropout(p)
    drop_path = None #StochasticDepth(p_path, mode="row")
    # run_naive(x, residual, drop_path, dropout, norm)

    torch.manual_seed(0)
    torch.cuda.manual_seed_all(0)
    norm = nn.LayerNorm(d).cuda()
    dropout = nn.Dropout(p)
    drop_path = None #StochasticDepth(p_path, mode="row")
    out, resid = run_torch(x, residual, drop_path, dropout, norm)

    outs = []
    resids = []
    for _name, fn in [('tk', run_tk), ('triton', run_flash)]:
        print(f"Running {_name}")
        torch.manual_seed(0)
        torch.cuda.manual_seed_all(0)
        norm = nn.LayerNorm(d).cuda()
        dropout = nn.Dropout(p)
        drop_path = None #StochasticDepth(p_path, mode="row")
        fn_out, fn_resid = fn(x, residual, drop_path, dropout, norm)

        print("----"*10)
        diff = torch.norm(out - fn_out).item()
        print(out[2,4,:8])
        print(fn_out[2,4,:8])
        print(f"Out Diff: {diff}")

        diff = torch.norm(resid - fn_resid).item()
        print(resid[4,2,:8])
        print(fn_resid[4,2,:8])
        print(f"Resid Diff: {diff}")

        print("----"*10)

        # breakpoint()





import torch
import thunderkittens as tk
import numpy as np
from tqdm import tqdm

from flash_attn_interface import flash_attn_func

def generate_tensor(shape, mean, std, dtype, device):
    tensor = torch.randn(shape, dtype=dtype, device=device)
    magnitude = torch.norm(tensor, dim=-1, keepdim=True)
    scaled_tensor = tensor * (torch.randn(magnitude.shape, dtype=dtype, device=device) * std + mean) / magnitude
    return scaled_tensor.contiguous()

def tk_forward_test(Q, K, V, causal):
    O = torch.zeros_like(Q).contiguous()
    L = torch.zeros(Q.shape[0], Q.shape[1], Q.shape[2], 1, device=Q.device, dtype=torch.float)
    tk.mha_forward(Q, K, V, O, L, causal)
    return O, L

def tk_backward_test(Q, K, V, O, L, d_vec, dO, causal):
    return tk.mha_backward(Q, K, V, O, L, d_vec, dO, causal)

def check_consistency(b, h, n, d, causal, mean, std, num_iterations=100000):
    # Generate fixed inputs
    torch.manual_seed(0)

    max_diff_O = 0
    max_diff_L = 0
    max_diff_qg, max_diff_kg, max_diff_vg = 0, 0, 0
    threshold = 1e-5
    for _ in tqdm(range(num_iterations), desc="Checking consistency"):
        Q  = generate_tensor((b, h, n, d), mean, std, torch.bfloat16, 'cuda')
        K  = generate_tensor((b, h, n, d), mean, std, torch.bfloat16, 'cuda')
        V  = generate_tensor((b, h, n, d), mean, std, torch.bfloat16, 'cuda')
        dO = generate_tensor((b, h, n, d), mean, std, torch.bfloat16, 'cuda')
        Q.requires_grad, K.requires_grad, V.requires_grad = True, True, True

        high_precision_Q, high_precision_K, high_precision_V = Q.clone().detach().to(torch.float64), K.clone().detach().to(torch.float64), V.clone().detach().to(torch.float64)
        high_precision_Q.requires_grad, high_precision_K.requires_grad, high_precision_V.requires_grad = True, True, True
        high_precision_attn = torch.einsum('bhnd,bhmd->bhnm', high_precision_Q, high_precision_K)
        high_precision_attn/=d**.5
        if causal:
            causal_mask = torch.tril(torch.ones_like(high_precision_attn), diagonal=0)
            high_precision_attn = high_precision_attn.masked_fill(causal_mask == 0, float('-inf'))
        high_precision_attn = torch.nn.functional.softmax(high_precision_attn, dim=-1)
        high_precision_o = torch.einsum('bhnm,bhmd->bhnd', high_precision_attn, high_precision_V).to(torch.bfloat16)
        high_precision_o.backward(dO)
        high_precision_qg = high_precision_Q.grad.to(torch.bfloat16)
        high_precision_kg = high_precision_K.grad.to(torch.bfloat16)
        high_precision_vg = high_precision_V.grad.to(torch.bfloat16)
        high_precision_Q.grad, high_precision_K.grad, high_precision_V.grad = None, None, None

        ref_O = torch.nn.functional.scaled_dot_product_attention(Q, K, V, is_causal=causal)
        ref_O.backward(dO)
        ref_qg = Q.grad.to(torch.bfloat16)
        ref_kg = K.grad.to(torch.bfloat16)
        ref_vg = V.grad.to(torch.bfloat16)
        Q.grad, K.grad, V.grad = None, None, None

        # Forward pass
        O, L = tk_forward_test(Q.clone().detach(), K.clone().detach(), V.clone().detach(), causal)
        
        # Backward pass
        d_vec = torch.zeros(b, h, n, 1, device='cuda', dtype=torch.float)
        qg, kg, vg = tk_backward_test(Q.clone().detach(), K.clone().detach(), V.clone().detach(), O.clone().detach(), L.clone().detach(), d_vec, dO, causal)
        
        # Check differences
        max_diff_O = torch.abs(O - high_precision_o).max().item()
        max_diff_qg = torch.abs(qg - high_precision_qg).max().item()
        max_diff_kg = torch.abs(kg - high_precision_kg).max().item()
        max_diff_vg = torch.abs(vg - high_precision_vg).max().item()
        
        if max_diff_O > threshold or max_diff_L > threshold or max_diff_qg > threshold or max_diff_kg > threshold or max_diff_vg > threshold:
            print(f"Iteration {_}: Large difference detected")
            print(f"Max diff O: {max_diff_O}, Avg diff O: {torch.abs(O - high_precision_o).mean().item()}")
            print(f"Max diff qg: {max_diff_qg}, Avg diff qg: {torch.abs(qg - high_precision_qg).mean().item()}")
            print(f"Max diff kg: {max_diff_kg}, Avg diff kg: {torch.abs(kg - high_precision_kg).mean().item()}")
            print(f"Max diff vg: {max_diff_vg}, Avg diff vg: {torch.abs(vg - high_precision_vg).mean().item()}")
            print(f"Max FA2 diff O: {torch.abs(ref_O - high_precision_o).max().item()}, Avg diff O: {torch.abs(ref_O - high_precision_o).mean().item()}")
            print(f"Max FA2 diff qg: {torch.abs(ref_qg - high_precision_qg).max().item()}, Avg diff qg: {torch.abs(ref_qg - high_precision_qg).mean().item()}")
            print(f"Max FA2 diff kg: {torch.abs(ref_kg - high_precision_kg).max().item()}, Avg diff kg: {torch.abs(ref_kg - high_precision_kg).mean().item()}")
            print(f"Max FA2 diff vg: {torch.abs(ref_vg - high_precision_vg).max().item()}, Avg diff vg: {torch.abs(ref_vg - high_precision_vg).mean().item()}")
            breakpoint()

    return max_diff_O, max_diff_L, max_diff_qg, max_diff_kg, max_diff_vg

# Example usage
b, h, n, d = 12, 12, 768, 64
causal = True
mean = 15
std = 200

max_diff_O, max_diff_L, max_diff_qg, max_diff_kg, max_diff_vg = check_consistency(b, h, n, d, causal, mean, std)

print(f"Maximum difference in O across iterations: {max_diff_O}")
print(f"Maximum difference in L across iterations: {max_diff_L}")
print(f"Maximum difference in qg across iterations: {max_diff_qg}")
print(f"Maximum difference in kg across iterations: {max_diff_kg}")
print(f"Maximum difference in vg across iterations: {max_diff_vg}")


# Copyright (c) 2023, Tri Dao.
# Adapted by Simran Arora and Sabri Eyuboglu.

import math
from functools import partial

import torch
import torch.nn as nn
from einops import rearrange, repeat

from flash_attn import (
    flash_attn_kvpacked_func,
    flash_attn_qkvpacked_func,
    flash_attn_varlen_kvpacked_func,
    flash_attn_varlen_qkvpacked_func,
    flash_attn_with_kvcache,
    flash_attn_func,
)

import inspect
_flash_supports_window_size = "window_size" in list(inspect.signature(flash_attn_func).parameters)
assert _flash_supports_window_size, "flash_attn_func does not support window_size"

try:
    from flash_attn.ops.fused_dense import ColumnParallelLinear, FusedDense, RowParallelLinear
except ImportError:
    FusedDense, ColumnParallelLinear, RowParallelLinear = None, None, None

try:
    from flash_attn.layers.rotary import RotaryEmbedding
except ImportError:
    RotaryEmbedding = None

from based.models.mixers.rotary import get_rotary_embeddings, apply_rotary_pos_emb

class FlashSelfAttention(nn.Module):
    """Implement the scaled dot product attention with softmax.
    Arguments
    ---------
        softmax_scale: The temperature to use for the softmax attention.
                      (default: 1/sqrt(d_keys) where d_keys is computed at
                      runtime)
        attention_dropout: The dropout rate to apply to the attention
                           (default: 0.0)
    """

    def __init__(self, causal=True, softmax_scale=None, attention_dropout=0.0, window_size=None):
        super().__init__()
        assert flash_attn_varlen_qkvpacked_func is not None, "FlashAttention is not installed"
        assert flash_attn_qkvpacked_func is not None, "FlashAttention is not installed"
        self.causal = causal
        self.softmax_scale = softmax_scale
        self.drop = nn.Dropout(attention_dropout)
        self.window_size = window_size

    def forward(self, qkv, causal=True, cu_seqlens=None, max_seqlen=None, **kwargs):
        """Implements the multihead softmax attention.
        Arguments
        ---------
            qkv: The tensor containing the query, key, and value.
                If cu_seqlens is None and max_seqlen is None, then qkv has shape (B, S, 3, H, D).
                If cu_seqlens is not None and max_seqlen is not None, then qkv has shape
                (total, 3, H, D), where total is the sum of the sequence lengths in the batch.
            causal: if passed, will override self.causal
            cu_seqlens: (batch_size + 1,), dtype torch.int32. The cumulative sequence lengths
                of the sequences in the batch, used to index into qkv.
            max_seqlen: int. Maximum sequence length in the batch.
        Returns:
        --------
            out: (total, H, D) if cu_seqlens is not None and max_seqlen is not None,
                else (B, S, H, D).
        """
        in_dtype = qkv.dtype
        if qkv.dtype not in [torch.float16, torch.bfloat16]:
            qkv = qkv.to(torch.bfloat16)

        assert qkv.dtype in [torch.float16, torch.bfloat16]
        assert qkv.is_cuda
        causal = True # Changed (01152023)
        unpadded = cu_seqlens is not None

        if self.window_size is not None:
            if unpadded:
                assert cu_seqlens.dtype == torch.int32
                assert max_seqlen is not None
                assert isinstance(max_seqlen, int)
                return flash_attn_varlen_qkvpacked_func(
                    qkv,
                    cu_seqlens,
                    max_seqlen,
                    self.drop.p if self.training else 0.0,
                    softmax_scale=self.softmax_scale,
                    causal=causal,
                    window_size=self.window_size ,    # [i - window_size[0], i + window_size[1]]
                ).to(in_dtype)
            else:
                return flash_attn_qkvpacked_func(
                    qkv,
                    self.drop.p if self.training else 0.0,
                    softmax_scale=self.softmax_scale,
                    causal=causal,
                    window_size=self.window_size,  
                ).to(in_dtype)
        else:
            assert 0, print("Using windows")
            if unpadded:
                assert cu_seqlens.dtype == torch.int32
                assert max_seqlen is not None
                assert isinstance(max_seqlen, int)
                return flash_attn_varlen_qkvpacked_func(
                    qkv,
                    cu_seqlens,
                    max_seqlen,
                    self.drop.p if self.training else 0.0,
                    softmax_scale=self.softmax_scale,
                    causal=causal,
                ).to(in_dtype)
            else:
                return flash_attn_qkvpacked_func(
                    qkv,
                    self.drop.p if self.training else 0.0,
                    softmax_scale=self.softmax_scale,
                    causal=causal,
                ).to(in_dtype)


class FlashCrossAttention(nn.Module):
    """Implement the scaled dot product attention with softmax.
    Arguments
    ---------
        softmax_scale: The temperature to use for the softmax attention.
                      (default: 1/sqrt(d_keys) where d_keys is computed at
                      runtime)
        attention_dropout: The dropout rate to apply to the attention
                           (default: 0.0)
    """

    def __init__(self, causal=True, softmax_scale=None, attention_dropout=0.0, window_size=None):
        super().__init__()
        assert flash_attn_varlen_kvpacked_func is not None, "FlashAttention is not installed"
        assert flash_attn_kvpacked_func is not None, "FlashAttention is not installed"
        self.causal = causal
        self.softmax_scale = softmax_scale
        self.drop = nn.Dropout(attention_dropout)
        self.window_size = window_size

    def forward(
        self,
        q,
        kv,
        causal=None,
        cu_seqlens=None,
        max_seqlen=None,
        cu_seqlens_k=None,
        max_seqlen_k=None,
    ):
        """Implements the multihead softmax attention.
        Arguments
        ---------
            q: The tensor containing the query. (B, Sq, H, D)
            kv: The tensor containing the key and value. (B, Sk, 2, H_k, D)
            causal: if passed, will override self.causal
            cu_seqlens: (batch_size + 1,), dtype torch.int32. The cumulative sequence lengths
                of the sequences in the batch, used to index into q.
            max_seqlen: int. Maximum sequence length in the batch of q.
            cu_seqlens_k: (batch_size + 1,), dtype torch.int32. The cumulative sequence lengths
                of the sequences in the batch, used to index into kv.
            max_seqlen_k: int. Maximum sequence length in the batch of k and v.
        """
        in_dtype = q.dtype
        if q.dtype not in [torch.float16, torch.bfloat16]:
            q = q.to(torch.bfloat16)
            kv = kv.to(torch.bfloat16)
        assert q.dtype in [torch.float16, torch.bfloat16]
        assert q.is_cuda and kv.is_cuda
        causal = True #self.causal if causal is None else causal
        unpadded = cu_seqlens is not None
        if self.window_size is not None:
            if unpadded:
                assert cu_seqlens.dtype == torch.int32
                assert max_seqlen is not None
                assert isinstance(max_seqlen, int)
                assert cu_seqlens_k is not None
                assert cu_seqlens_k.dtype == torch.int32
                assert max_seqlen_k is not None
                assert isinstance(max_seqlen, int)
                return flash_attn_varlen_kvpacked_func(
                    q,
                    kv,
                    cu_seqlens,
                    cu_seqlens_k,
                    max_seqlen,
                    max_seqlen_k,
                    self.drop.p if self.training else 0.0,
                    softmax_scale=self.softmax_scale,
                    causal=True,
                    window_size=self.window_size,
                ).to(in_dtype)
            else:
                batch_size, seqlen_q = q.shape[0], q.shape[1]
                seqlen_k = kv.shape[1]
                assert kv.shape[0] == batch_size and kv.shape[4] == q.shape[3]
                return flash_attn_kvpacked_func(
                    q,
                    kv.to(q.dtype),
                    self.drop.p if self.training else 0.0,
                    causal=True,
                    softmax_scale=self.softmax_scale,
                    window_size=self.window_size,
                ).to(in_dtype)
        else:
            assert 0, print("Using windows")
            if unpadded:
                assert cu_seqlens.dtype == torch.int32
                assert max_seqlen is not None
                assert isinstance(max_seqlen, int)
                assert cu_seqlens_k is not None
                assert cu_seqlens_k.dtype == torch.int32
                assert max_seqlen_k is not None
                assert isinstance(max_seqlen, int)
                return flash_attn_varlen_kvpacked_func(
                    q,
                    kv,
                    cu_seqlens,
                    cu_seqlens_k,
                    max_seqlen,
                    max_seqlen_k,
                    self.drop.p if self.training else 0.0,
                    softmax_scale=self.softmax_scale,
                    causal=True,
                ).to(in_dtype)
            else:
                batch_size, seqlen_q = q.shape[0], q.shape[1]
                seqlen_k = kv.shape[1]
                assert kv.shape[0] == batch_size and kv.shape[4] == q.shape[3]
                return flash_attn_kvpacked_func(
                    q,
                    kv,
                    self.drop.p if self.training else 0.0,
                    causal=True,
                    softmax_scale=self.softmax_scale,
                ).to(in_dtype)


class LinearResidual(nn.Linear):
    """Wrap nn.Linear to return the residual as well. For compatibility with FusedDense."""

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return super().forward(input), input


def _update_kv_cache(kv, inference_params, layer_idx, inference_mode='default', window_size=None):
    """kv: (batch_size, seqlen, 2, nheads, head_dim) or (batch_size, 1, 2, nheads, head_dim)"""

    if inference_mode == "default" or inference_mode == "default_rotary":
        # Pre-allocate memory for key-values for inference.
        num_heads, head_dim = kv.shape[-2:]
        if layer_idx not in inference_params.key_value_memory_dict:
            kv_cache = torch.empty(
                inference_params.max_batch_size,
                inference_params.max_seqlen,
                2,
                num_heads,
                head_dim,
                dtype=kv.dtype,
                device=kv.device,
            )
            inference_params.key_value_memory_dict[layer_idx] = kv_cache
        else:
            kv_cache = inference_params.key_value_memory_dict[layer_idx]
        # Adjust key and value for inference
        batch_start = inference_params.batch_size_offset
        batch_end = batch_start + kv.shape[0]
        sequence_start = inference_params.seqlen_offset
        sequence_end = sequence_start + kv.shape[1]
        assert batch_end <= (kv_cache.shape[0] if kv_cache is not None else kv_cache.shape[0])
        assert sequence_end <= (kv_cache.shape[1] if kv_cache is not None else kv_cache.shape[2])
        assert kv_cache is not None
        kv_cache[batch_start:batch_end, sequence_start:sequence_end, ...] = kv

        return kv_cache[batch_start:batch_end, :sequence_end, ...], None, None
    else:
        assert window_size is not None, print("Implementing this for SWA.")

        cache_size = min(inference_params.max_seqlen, window_size) 
        # Pre-allocate memory for key-values for inference.
        num_heads, head_dim = kv.shape[-2:]
        if layer_idx not in inference_params.key_value_memory_dict:
            kv_cache = torch.empty(
                inference_params.max_batch_size,
                cache_size,    # SA: Change 1
                2,
                num_heads,
                head_dim,
                dtype=kv.dtype,
                device=kv.device,
            )
            inference_params.key_value_memory_dict[layer_idx] = kv_cache
        else:
            kv_cache = inference_params.key_value_memory_dict[layer_idx]

        # Adjust key and value for inference
        batch_start = inference_params.batch_size_offset
        batch_end = batch_start + kv.shape[0]
        sequence_start = inference_params.seqlen_offset
        sequence_end = sequence_start + kv.shape[1]
        
        assert batch_end <= (kv_cache.shape[0] if kv_cache is not None else kv_cache.shape[0])
        # assert sequence_end <= (kv_cache.shape[1] if kv_cache is not None else kv_cache.shape[2]) # SA: Change 2
        assert kv_cache is not None
        kv_cache[batch_start:batch_end, :min(kv.shape[1], cache_size), ...] = kv[:, max(0,(kv.shape[1]-cache_size)):,:,:,:] # SA: Change 3]
        return kv_cache[batch_start:batch_end, :sequence_end, ...], cache_size, kv


class SlidingAttention(nn.Module):
    """Multi-head self-attention and cross-attention"""

    def __init__(
        self,
        embed_dim,
        num_heads,
        num_heads_kv=None,
        cross_attn=False,
        qkv_proj_bias=True,
        out_proj_bias=True,
        dropout=0.0,
        softmax_scale=None,
        causal=True,
        layer_idx=None,
        rotary_emb_dim=0,
        rotary_emb_base=10000.0,
        rotary_emb_scale_base=None,
        rotary_emb_interleaved=False,
        fused_bias_fc=False,
        use_flash_attn=True,
        return_residual=False,
        checkpointing=False,
        window_size=128,
        l_max=None,
        inference_bs=1,
        inference_mode="default",
        device=None,
        dtype=None,
    ) -> None:
        """
        num_heads_kv: can be used to toggle MQA / GQA. If None, use num_heads.
        return_residual: whether to return the input x along with the output. This is for
            performance reason: for post-norm architecture, returning the input allows us
            to fuse the backward of nn.Linear with the residual connection.
        """
        # SA hardcode
        use_flash_attn = True
        causal = True

        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.embed_dim = embed_dim
        self.cross_attn = cross_attn
        self.causal = causal
        self.layer_idx = layer_idx
        self.rotary_emb_dim = rotary_emb_dim
        self.use_flash_attn = use_flash_attn
        self.return_residual = return_residual
        self.checkpointing = checkpointing

        if window_size is None:
            assert 0, print("Using windows")
            self.window = None
        else:
            self.window = (window_size//2, 0)
        self.window_size = self.window[0]

        # inference settings
        self.inference_bs = inference_bs
        self.inference_mode= inference_mode# "default_rotary" # SA: Flag 2
        self.cache_size = self.window_size + 1

        self.num_heads = num_heads
        self.num_heads_kv = num_heads_kv if num_heads_kv is not None else num_heads
        assert (self.num_heads % self.num_heads_kv == 0), "num_heads must be divisible by num_heads_kv"
        assert self.embed_dim % num_heads == 0, "embed_dim must be divisible by num_heads"
        self.head_dim = self.embed_dim // num_heads
        qkv_dim = self.head_dim * (self.num_heads + 2 * self.num_heads_kv)
        kv_dim = 2 * self.head_dim * self.num_heads_kv

        if self.rotary_emb_dim > 0:
            assert not cross_attn, "MHA with rotary embedding does not support cross-attention yet"
            assert RotaryEmbedding is not None, "rotary_emb is not installed"
            if self.inference_mode == "default":
                self.rotary_emb = RotaryEmbedding(
                    self.rotary_emb_dim,
                    base=rotary_emb_base,
                    scale_base=rotary_emb_scale_base,
                    interleaved=rotary_emb_interleaved,
                    device=device,
                )
            else:
                position_ids = torch.arange(l_max).to(device)
                self.position_ids = position_ids.unsqueeze(0).repeat(inference_bs, 1)
                # Rotary embeddings
                self.rotary_emb = get_rotary_embeddings(
                    rope_scaling_type=None,
                    head_dim=self.rotary_emb_dim,
                    max_position_embeddings=l_max,
                    rope_theta=rotary_emb_base,
                    rope_scaling_factor=rotary_emb_scale_base, 
                    device=device,
                )

        # if fused_bias_fc and FusedDense is None: raise ImportError("fused_dense is not installed")
        if FusedDense is None:
            fused_bias_fc = False

        linear_cls = nn.Linear if not fused_bias_fc else FusedDense
        linear_resid_cls = (
            LinearResidual if not fused_bias_fc else partial(FusedDense, return_residual=True)
        )
        wqkv_cls = linear_cls if not self.return_residual else linear_resid_cls
        inner_attn_cls = FlashSelfAttention
        inner_cross_attn_cls = FlashCrossAttention 
        self.Wqkv = wqkv_cls(embed_dim, qkv_dim, bias=qkv_proj_bias, **factory_kwargs)
        self.inner_attn = inner_attn_cls(
            causal=causal, softmax_scale=softmax_scale, attention_dropout=dropout, window_size=self.window
        )
        self.inner_cross_attn = inner_cross_attn_cls(
            causal=causal, softmax_scale=softmax_scale, attention_dropout=dropout, window_size=self.window
        )
        self.out_proj = linear_cls(embed_dim, embed_dim, bias=out_proj_bias, **factory_kwargs)

    def allocate_inference_cache(self, batch_size, max_seqlen, dtype=None):
        dtype = self.out_proj.weight.dtype if dtype is None else dtype
        device = self.out_proj.weight.device
        if self.inference_mode == "default" or self.inference_mode == "default_rotary": 
            return torch.empty(batch_size,max_seqlen,2,self.num_heads_kv,self.head_dim,dtype=dtype,device=device,)
        else:
            return torch.empty(batch_size,self.window_size,2,self.num_heads_kv,self.head_dim,dtype=dtype,device=device,)

    def _update_kv_cache(self, kv, inference_params):
        """kv: (batch_size, seqlen, 2, nheads, head_dim) or (batch_size, 1, 2, nheads, head_dim)"""
        assert self.layer_idx is not None, "Generation requires layer_idx in the constructor"
        kv_cache, cache_size, cur_kv =  _update_kv_cache(kv, inference_params, self.layer_idx, inference_mode = self.inference_mode, 
        window_size=self.window_size)
        self.cache_size = cache_size
        return kv_cache, cur_kv

    def _apply_rotary_update_kvcache_attention(self, q, kv, inference_params):
        """
        Fast path that combine 3 steps: apply rotary to Q and K, update kv cache, and apply attention.
        q: (batch_size, seqlen_q, nheads, head_dim)
        kv: (batch_size, seqlen_k, 2, nheads_kv, head_dim)
        """
        assert inference_params is not None and inference_params.seqlen_offset > 0
        assert self.use_flash_attn, "This code path only supports FlashAttention"
        batch = q.shape[0]
        kv_cache = inference_params.key_value_memory_dict[self.layer_idx][:batch]
        in_dtype = q.dtype
        if q.dtype not in [torch.float16, torch.bfloat16]:
            q = q.to(torch.bfloat16)
            kv_cache = kv_cache.to(torch.bfloat16)
            kv = kv.to(torch.bfloat16)

        if self.rotary_emb_dim > 0:
            assert self.rotary_emb.scale is None, "This code path does not support xPos"
            self.rotary_emb._update_cos_sin_cache(
                inference_params.max_seqlen, device=q.device, dtype=q.dtype
            )
            rotary_cos, rotary_sin = self.rotary_emb._cos_cached, self.rotary_emb._sin_cached
        else:
            rotary_cos, rotary_sin = None, None

        cache_seqlens = (
            inference_params.lengths_per_sample[:batch]
            if inference_params.lengths_per_sample is not None
            else inference_params.seqlen_offset
        )

        context = flash_attn_with_kvcache(
            q,
            kv_cache[:, :, 0].to(q.dtype),
            kv_cache[:, :, 1].to(q.dtype),
            kv[:, :, 0],
            kv[:, :, 1],
            rotary_cos=rotary_cos,
            rotary_sin=rotary_sin,
            cache_seqlens=cache_seqlens,
            softmax_scale=self.inner_cross_attn.softmax_scale,
            causal=self.inner_cross_attn.causal,
            rotary_interleaved=self.rotary_emb.interleaved if self.rotary_emb_dim > 0 else False,
            window_size=self.window,
        ).to(in_dtype)
        return context

    def _update_kvcache_attention(self, q, kv, inference_params):
        """Write kv to inference_params, then do attention"""
        if (
            inference_params.seqlen_offset == 0
            or flash_attn_with_kvcache is None
            or not self.use_flash_attn
        ):
            # TODO: this only uses seqlen_offset and not lengths_per_sample.
            kv_cache, cur_kv = self._update_kv_cache(kv, inference_params)
            if self.inference_mode == "default" or self.inference_mode == "default_rotary": 
                return self.inner_cross_attn(q, kv_cache)
            else:
                return self.inner_cross_attn(q, cur_kv)
        else:
            batch = q.shape[0]
            kv_cache = inference_params.key_value_memory_dict[self.layer_idx][:batch]
            cache_seqlens = (
                # inference_params.lengths_per_sample[:batch]
                # if inference_params.lengths_per_sample is not None
                # else 
                inference_params.seqlen_offset
            )
            # breakpoint()
            in_dtype = q.dtype
            if q.dtype not in [torch.float16, torch.bfloat16]:
                q = q.to(torch.float16)
                kv_cache = kv_cache.to(torch.float16)
                kv = kv.to(torch.float16)
            if self.window is None:
                assert 0, print("Using windows")
                return flash_attn_with_kvcache(
                    q,
                    kv_cache[:, :, 0],
                    kv_cache[:, :, 1],
                    kv[:, :, 0],
                    kv[:, :, 1],
                    cache_seqlens=cache_seqlens,
                    softmax_scale=self.inner_cross_attn.softmax_scale,
                    causal=self.inner_cross_attn.causal,
                ).to(in_dtype)
            else:
                
                if self.inference_mode == "default" or self.inference_mode == "default_rotary" or cache_seqlens < self.cache_size:
                    length = cache_seqlens
                else:
                    length = self.cache_size-1

                # circular buffering
                if self.cache_size and self.cache_size < cache_seqlens:
                    kv_cache = torch.roll(kv_cache, -1, dims=1) # make room at the end
                    kv_cache[:,-1,:,:,:] = torch.zeros_like(kv_cache[:,-1,:,:,:])   # clear the end

                out = flash_attn_with_kvcache(
                    q,
                    kv_cache[:, :, 0],
                    kv_cache[:, :, 1],
                    kv[:, :, 0],
                    kv[:, :, 1],
                    cache_seqlens=length,
                    softmax_scale=self.inner_cross_attn.softmax_scale,
                    causal=self.inner_cross_attn.causal,
                    window_size=self.window,
                ).to(in_dtype)
                return out

    def forward(
        self,
        x,
        x_kv=None,
        key_padding_mask=None,
        cu_seqlens=None,
        max_seqlen=None,
        mixer_subset=None,
        inference_params=None,
        **kwargs,
    ):
        """
        Arguments:
            x: (batch, seqlen, hidden_dim) (where hidden_dim = num heads * head dim) if
                cu_seqlens is None and max_seqlen is None, else (total, hidden_dim) where total
                is the is the sum of the sequence lengths in the batch.
            x_kv: (batch, seqlen, hidden_dim), only applicable for cross-attention. If None, use x.
            cu_seqlens: (batch_size + 1,), dtype torch.int32. The cumulative sequence lengths
                of the sequences in the batch, used to index into x. Only applicable when using
                FlashAttention.
            max_seqlen: int. Maximum sequence length in the batch.
            key_padding_mask: boolean mask, True means to keep, False means to mask out.
                (batch, seqlen). Only applicable when not using FlashAttention.
            mixer_subset: for cross-attention only. If not None, will take a subset of x
                before applying the query projection. Useful for e.g., ViT where we only care
                about the CLS token in the last layer.
            inference_params: for generation. Adapted from Megatron-LM (and Apex)
            https://github.com/NVIDIA/apex/blob/3ff1a10f72ec07067c4e44759442329804ac5162/apex/transformer/testing/standalone_transformer_lm.py#L470
        """

        if cu_seqlens is not None:
            assert max_seqlen is not None
            assert key_padding_mask is None
            assert self.use_flash_attn
            assert self.rotary_emb_dim == 0
        if key_padding_mask is not None:
            assert cu_seqlens is None
            assert max_seqlen is None
            assert not self.use_flash_attn
        if inference_params is not None:
            assert key_padding_mask is None
            assert cu_seqlens is None and max_seqlen is None

        kwargs = (
            {"cu_seqlens": cu_seqlens, "max_seqlen": max_seqlen, **kwargs}
            if self.use_flash_attn
            else {"key_padding_mask": key_padding_mask, **kwargs}
        )
        seqlen_offset = (
            0
            if inference_params is None
            else (
                inference_params.lengths_per_sample
                if inference_params.lengths_per_sample is not None
                else inference_params.seqlen_offset
            )
        )
        rotary_max_seqlen = inference_params.max_seqlen if inference_params is not None else None
        batch, seqlen = x.shape[:2]
        if not self.cross_attn and self.num_heads_kv == self.num_heads:
            assert x_kv is None and mixer_subset is None
            if not self.return_residual: qkv = self.Wqkv(x)
            else: qkv, x = self.Wqkv(x)
            qkv = rearrange(qkv, "... (three h d) -> ... three h d", three=3, d=self.head_dim)
            if (
                inference_params is None
                or inference_params.seqlen_offset == 0
                or (self.rotary_emb_dim == 0 or self.rotary_emb_dim % 16 != 0)
                or not self.use_flash_attn or 
                self.inference_mode != "default"
            ):
                if self.rotary_emb_dim > 0: 
                    if self.inference_mode == "default":

                        qkv = self.rotary_emb(
                            qkv, seqlen_offset=seqlen_offset, max_seqlen=rotary_max_seqlen
                        )
                    else:
                        cos, sin = self.rotary_emb(qkv, seq_len=rotary_max_seqlen)  
                        q, k = apply_rotary_pos_emb(
                            qkv[:,:,0,:,:].transpose(1,2), 
                            qkv[:,:,1,:,:].transpose(1,2), 
                            cos, sin, self.position_ids[:,:seqlen]
                        )
                        qkv = torch.concat([q.transpose(1,2).unsqueeze(2),k.transpose(1,2).unsqueeze(2), qkv[:,:,2,:,:].unsqueeze(2),], dim=2)

                if inference_params is None:
                    if not self.checkpointing:
                        context = self.inner_attn(qkv, **kwargs)
                    else:
                        context = torch.utils.checkpoint.checkpoint(self.inner_attn, qkv, **kwargs)
                else:
                    context = self._update_kvcache_attention(
                        qkv[:, :, 0], qkv[:, :, 1:], inference_params
                    )
            else:
                context = self._apply_rotary_update_kvcache_attention(
                    qkv[:, :, 0], qkv[:, :, 1:], inference_params
                )
        else:
            assert self.num_heads_kv != self.num_heads
            if not self.return_residual: qkv = self.Wqkv(x)
            else: qkv, x = self.Wqkv(x)
            q = qkv[..., : self.num_heads * self.head_dim]
            kv = qkv[..., self.num_heads * self.head_dim :]
            q = rearrange(q, "... (h d) -> ... h d", d=self.head_dim)
            kv = rearrange(kv, "... (two hkv d) -> ... two hkv d", two=2, d=self.head_dim)
            if (
                inference_params is None
                or inference_params.seqlen_offset == 0
                or (self.rotary_emb_dim == 0 or self.rotary_emb_dim % 16 != 0)
                or not self.use_flash_attn
            ):
                if self.rotary_emb_dim > 0:
                    q, kv = self.rotary_emb(
                        q, kv, seqlen_offset=seqlen_offset, max_seqlen=rotary_max_seqlen
                    )
                if inference_params is None:
                    if not self.checkpointing:
                        context = self.inner_cross_attn(q, kv, **kwargs)
                    else:
                        context = torch.utils.checkpoint.checkpoint(
                            self.inner_cross_attn, q, kv, **kwargs
                        )
                else:
                    context = self._update_kvcache_attention(q, kv, inference_params)
            else:
                context = self._apply_rotary_update_kvcache_attention(q, kv, inference_params)
        out = self.out_proj(rearrange(context, "... h d -> ... (h d)"))
        return out if not self.return_residual else (out, x)




"""
LoLCATs + ThunderKittens linear attention + sliding window for generation
"""
from typing import Optional, Tuple, List
import torch
import torch.nn.functional as F

try:
    import thunderkittens as tk
    print(f"Successfully imported ThunderKittens for TK window attention")
except:
    print(f"Failed to import ThunderKittens for TK window attention")

from .linear_window_attention_tk_long import LolcatsTKWindowLongAttention
from .linear_attention import LinearAttentionState

class LolcatsWindowAttentionTKGen(LolcatsTKWindowLongAttention):
    def __init__(self, *args, window_size: int = 64, layer_idx=0, **kwargs):
        super().__init__(*args, **kwargs)
        self.train_attention = False
        self.base_inference = False 
        self.window_size = 64   # hard-coded support for TK kernel
        self.decode_window_size = 64
        self.layer_idx = layer_idx
        
        if self.layer_idx == 0: 
            print(f"Using LolcatsWindowAttentionTKGen!")
    
    def forward(self,
                hidden_states: torch.Tensor,
                attention_mask: Optional[torch.Tensor] = None,
                position_ids: Optional[torch.LongTensor] = None,
                past_key_value: Optional[Tuple[int, torch.Tensor, torch.Tensor]] = None,  # “legacy” cache approach
                output_attentions: bool = False,
                use_cache: bool = False,
                **kwargs,
               ) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[Tuple[torch.Tensor]]]:
        """
        Forward pass with the option to compute attention weights multiple ways
        if self.train_attention is True
        -> Consistent with HuggingFace Transformers for easy use with their pretrained models
        """
        b, l, _ = hidden_states.size()
        assert past_key_value is not None, "past_key_value must be provided for generation"
        assert self.train_attention is False, "train_attention is not supported for generation"
        assert self.base_inference is False, "base_inference is not supported for generation"
        assert use_cache is True, "use_cache must be True for generation"
        past_key_value.window_size = self.decode_window_size
        q, k, v, kv_seq_len = self.process_qkv(hidden_states, attention_mask, 
                                               position_ids, past_key_value)
        if q.shape[2] == 1 and kv_seq_len > 1:  
            # Inference generating after prefill
            f_q = self.feature_map_q(q)
            _kv = past_key_value.update_for_decoding(k, v, self.layer_idx,
                                                     self.feature_map_k)
            k_cache, v_cache, kv_state, k_state = _kv
            # Sliding window + linear attention decode
            window_factors = F.sigmoid(self.window_factors)
            linear_factors = 1 - window_factors if self.affine_attention_factors else 1

            # Softmax attention terms
            a_sm = torch.einsum('bhmd,bhnd->bhmn', q.float(), k_cache.float()) * (k.shape[-1] ** -0.5)
            a_sm_max = torch.amax(a_sm, dim=-1, keepdim=True)
            a_sm   = window_factors * torch.exp(a_sm - a_sm_max)
            sum_sm = a_sm.sum(dim=-1, keepdim=True)

            # Combine with linear attention terms
            y_true = (torch.einsum('bhmn,bhnd->bhmd', a_sm, v_cache.float())
                      + linear_factors * torch.einsum('bhld,bhdf->bhlf', f_q.float(), kv_state.float()))
            sum_ln = linear_factors * torch.einsum(
                'bhld,bhnd->bhl', f_q.float(), k_state.float())[..., None]
            y_true = (y_true / (sum_sm + sum_ln)).to(q.dtype) 
        
        else:  
            # Inference prefill 
            # Use TK-implemented linear + terrace window attention
            b, h, l, d = q.shape
            device = q.device
            # tk.hedgehog arguments
            betas    = F.sigmoid(self.window_factors[0, :, 0, 0].to(dtype=torch.float32))
            alphas   = (1 - betas if self.affine_attention_factors else 
                        torch.ones(betas.shape, dtype=torch.float32, device=device))
            q_map = self.feature_map_q.mlp.layer
            k_map = self.feature_map_k.mlp.layer
            # Saves outputs to y_pred, k_state, kv_state, where we fuse:
            # 1. f_q, f_k = self.feature_map_q(q), self.feature_map_k(k)
            # 2. y_pred = attention(q, k, f_q, f_k, v)  # b, h, l, d
            # 3. kv_state = torch.einsum(‘bhlf,bhld->bhfd’, 
            #                            f_k[:, :, :-self.window_size], 
            #                            v[:, :, :-self.window_size])  # b, h, f, d
            # 4. k_state = f_k[:, :, :-self.window_size].sum(dim=-2)   # b, h, d
            q = q.contiguous()
            k = k.contiguous()
            v = v.contiguous()
            y_true, kv_state, k_state = tk.hedgehog(q, k, v, q_map, k_map, alphas, betas)
            y_true = y_true.to(q.dtype)
            past_key_value.update_with_kv(kv_state, k_state, k, v, self.layer_idx)
        
        # Concatenate heads and apply output projection
        y_true = y_true.transpose(1, 2).contiguous().view(b, l, self.hidden_size)
        y_true = self.o_proj(y_true)
        return y_true, None, past_key_value


class LinearAttentionTKWindowGenerationCache(LinearAttentionState):
    """
    Class for `past_key_values`
    -> Alternative to KV cache; here we only maintain a “KV state” and “K state”
    -> Modified from transformers.cache_utils.DynamicCache (v4.36)  
    """
    def __init__(self, window_size: int = 64) -> None:
        super().__init__()
        self._seen_tokens = 0  # should be `self.seen_tokens` in Transformers v4.36
        self._seen_tokens_by_layer: List[int] = []
        self.window_size = window_size
        
        self.decode_kv_states: List[torch.Tensor] = []
        self.decode_k_states: List[torch.Tensor] = []
        self.k_cache: List[torch.Tensor] = []
        self.v_cache: List[torch.Tensor] = []

    def update_with_kv(self, 
                       kv_state: torch.Tensor, k_state: torch.Tensor,
                       k: torch.Tensor, v: torch.Tensor,
                       layer_idx: int) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Update the cache with new KV and K states
        """
        if layer_idx == 0:
            self._seen_tokens += k.shape[2]
        self._seen_tokens_by_layer.append(k.shape[2])

        # Initialize KV and K states
        if len(self.decode_k_states) <= layer_idx:  
            self.decode_kv_states.append(kv_state)
            self.decode_k_states.append(k_state)
        else:  # Update KV and K states
            self.decode_kv_states[layer_idx] = self.decode_kv_states[layer_idx] + kv_state
            self.decode_k_states[layer_idx] = self.decode_k_states[layer_idx] + k_state

        self.k_cache.append(k[:, :, -self.window_size:, :])
        self.v_cache.append(v[:, :, -self.window_size:, :])

    def update_for_decoding(self, k: torch.Tensor, v: torch.Tensor, 
                            layer_idx: int, feature_map_k: callable) -> None:
        """
        Update the cache for decoding
        """
        k_cache = self.k_cache[layer_idx]
        v_cache = self.v_cache[layer_idx]
        k_state = feature_map_k(k_cache[:, :, :1, :])
        v_state = v_cache[:, :, :1, :]
        kv_state = torch.einsum('bhlf,bhld->bhfd', k_state.float(), v_state.float()).to(k.dtype)

        self.decode_kv_states[layer_idx] += kv_state
        self.decode_k_states[layer_idx] += k_state

        self.k_cache[layer_idx] = torch.cat([k_cache[:, :, 1:, :], k], dim=-2)
        self.v_cache[layer_idx] = torch.cat([v_cache[:, :, 1:, :], v], dim=-2)
        if layer_idx == 0:
            self._seen_tokens += k.shape[-2]
        self._seen_tokens_by_layer[layer_idx] += k.shape[-2]
        return (self.k_cache[layer_idx], self.v_cache[layer_idx],
                self.decode_kv_states[layer_idx], self.decode_k_states[layer_idx])
    
    


"""
Learnable linear attention feature map classes and functions
"""
import torch
import torch.nn as nn
import torch.nn.functional as F


def init_feature_map(name: str, mlp: nn.Module, **kwargs: dict):
    """
    Initialize feature map final activation for linear attention
    """
    return FeatureMap(activation_name=name, mlp=mlp, **kwargs)


def init_feature_map_act(name: str, fullspace: bool = True, **kwargs):
    """
    Initialize feature map final activation for linear attention
    """
    if name == 'softmax_dim' and fullspace:
        return SoftmaxDim(**kwargs)
    elif name == 'softmax_dim' and not fullspace:
        return SoftmaxDimHalfspace(**kwargs)
    elif name == 'exp_dim' and fullspace:
        return Exp(**kwargs)
    elif name == 'exp_dim' and not fullspace:
        return ExpHalfspace(**kwargs)
    elif name == 'pos_elu':
        return PosELU(**kwargs)
    elif name == 'relu':
        return ReLU(**kwargs)
    
    else:
        raise NotImplementedError


def init_learned_kernel(name: str, **kwargs: any):
    """
    Initialize feature map MLP for linear attention
    """
    if name == 'untied_head_einsum':
        return FeatureMapMLP(**kwargs)
    elif name == 'untied_head_adapter':
        return FeatureMapAdapter(**kwargs)
    else:
        raise NotImplementedError


class FeatureMap(nn.Module):
    """
    Final 'activation' of feature map. Can probably be combined with
    `FeatureMapMLP` below

    Full feature map is like f(xW + b)
    -> This is the `f` part
    """
    def __init__(self,
                 activation_name: str,
                 head_dim_idx: int = -1, 
                 eps: float = 1e-12, 
                 mlp: nn.Module = None,
                 fullspace: bool = True,):
        super().__init__()
        self.head_dim_idx = head_dim_idx     
        self.eps = eps
        self.mlp = mlp if mlp is not None else nn.Identity()
        self.activation = init_feature_map_act(activation_name, fullspace, eps=eps)
        
    def forward(self, x: torch.Tensor, *mlp_args: any, **mlp_kwargs: any):
        """
        Assume x.shape is (batch_size, n_heads, seq_len, head_dim)
        """
        return self.activation(self.mlp(x, *mlp_args, **mlp_kwargs), x)

    def q_map(self, *args: any, **kwargs: any):
        """
        Use for inference in case q and k feature maps differ
        """
        return self.forward(*args, **kwargs)

    def k_map(self, *args: any, **kwargs: any):
        """
        Use for inference in case q and k feature maps differ
        """
        return self.forward(*args, **kwargs)


# -----------------------
# Feature map activations
# -----------------------
class FeatureMapAct(nn.Module):
    """
    Base class for feature map activations
    """
    def __init__(self, eps: float = 1e-12):
        super().__init__()
        self.eps = eps

    def forward(self, x: torch.Tensor, *args: any, **kwargs: any):
        """
        x.shape is (batch_size, n_heads, seq_len, head_dim)
        """
        return x


class PosELU(FeatureMapAct):
    """
    1 + ELU activation as in https://arxiv.org/abs/2006.16236
    """
    def forward(self, x: torch.Tensor, *args: any, **kwargs: any):
        return (1 + F.elu(x)).clamp(min=self.eps)


class ReLU(FeatureMapAct):
    """
    ReLU activation as in https://arxiv.org/abs/2103.13076
    """
    def forward(self, x: torch.Tensor, *args: any, **kwargs: any):
        return F.relu(x).clamp(min=self.eps)


class SoftmaxDim(FeatureMapAct):
    """
    Softmax activation as in https://arxiv.org/abs/2402.04347
    """
    def forward(self, x: torch.Tensor, *args: any, **kwargs: any):
        return torch.cat([
            torch.softmax(x, dim=-1), torch.softmax(-x, dim=-1)
        ], dim=-1).clamp(min=self.eps)

                         
class SoftmaxDimHalfspace(FeatureMapAct):
    """
    Softmax activation as in https://arxiv.org/abs/2402.04347
    """
    def forward(self, x: torch.Tensor, *args: any, **kwargs: any):
        return torch.softmax(x, dim=-1).clamp(min=self.eps)


class Exp(FeatureMapAct):
    """
    Exp activation as in https://arxiv.org/abs/2402.04347
    """
    def forward(self, x: torch.Tensor, *args: any, **kwargs: any):
        x_max = torch.amax(x, dim=-1, keepdim=True)
        x_min = torch.amin(x, dim=-1, keepdim=True)
        return torch.cat([
            torch.exp(x - x_max), torch.exp(-x + x_min)
        ], dim=-1).clamp(min=self.eps)


class ExpHalfspace(FeatureMapAct):
    """
    Exp activation as in https://arxiv.org/abs/2402.04347
    """
    def forward(self, x: torch.Tensor, *args: any, **kwargs: any):
        x_max = torch.amax(x, dim=-1, keepdim=True)
        return torch.exp(x - x_max).clamp(min=self.eps)


# ----------------
# Feature map MLPs
# ----------------

class FeatureMapMLP(nn.Module):
    """
    Learnable MLP in feature map.
    
    Full feature map is like f(xW + b)
    -> This is the `W` and (optional) `b` part
    """
    def __init__(self, 
                 num_heads: int,
                 head_dim: int,     # input dim
                 feature_dim: int,  # output dim
                 dtype: torch.dtype,
                 device: torch.device,
                 skip_connection: bool = False,
                 bias: bool = False,
                 zero_init: bool = False,
                 normal_init: bool = False,):
        super().__init__()
        self.num_heads = num_heads
        self.head_dim = head_dim
        self.feature_dim = feature_dim
        self.dtype = dtype
        self.device = device
        self.skip_connection = skip_connection
        self.bias = bias
        self.zero_init = zero_init
        self.normal_init = normal_init
        self.init_weights_()

        if self.zero_init:  # Zero-out weights or set as identity post-initialization
            self.zero_init_with_skip_() if self.skip_connection else self.zero_init_()

        if self.normal_init:
            with torch.no_grad():
                nn.init.normal_(self.layer)
        
        if self.skip_connection:
            assertion_fail = f'If self.skip_connection we need self.head_dim == self.feature_dim but self.head_dim is {self.head_dim} != self.feature_dim is {self.feature_dim}'
            assert self.head_dim == self.feature_dim, assertion_fail

    def init_weights_(self):
        """
        Initialize (W)eights and (b)iases
        """
        self.layer = nn.Parameter(torch.zeros(
            (self.num_heads, self.head_dim, self.feature_dim),
            dtype=self.dtype, device=self.device,
        ))
        nn.init.kaiming_uniform_(self.layer)

        if self.bias:
            self.bias = nn.Parameter(torch.zeros(
                (1, self.num_heads, 1, 1),  # self.feature_dim),
                dtype=self.dtype, device=self.device,
            ))
            nn.init.kaiming_uniform_(self.bias)
        else:
            self.bias = 0.  # hack

    def zero_init_with_skip_(self):
        """
        Initialize weights to zero matrix if skip connection
        """
        with torch.no_grad():
            nn.init.zeros_(self.layer)

    def zero_init_(self):
        """
        Initialize weights to identity matrix if no skip connection
        """
        with torch.no_grad():
            for i in range(self.layer.shape[0]):
                try:
                    nn.init.eye_(self.layer[i])
                except RuntimeError:
                    with torch.no_grad():
                        dtype = self.layer[i].dtype
                        weight = torch.eye(*self.layer[i].shape,
                                           requires_grad=self.layer[i].requires_grad,
                                           device=self.layer[i].device)
                        self.layer[i] = weight.to(dtype=dtype)

    def forward(self, x: torch.Tensor):
        """
        Assume x.shape is (batch_size, num_heads, seq_len, head_dim)
        """
        _x = torch.einsum('hdf,bhld->bhlf', self.layer, x) + self.bias
        return x + _x if self.skip_connection else _x


class FeatureMapAdapter(FeatureMapMLP):
    """
    Learnable Feature map with bottleneck adapter
    as in https://arxiv.org/abs/1902.00751

    We don't use but could be fun to try
    """
    def __init__(self, hidden_dim: int, *args, **kwargs):
        kwargs['skip_connection'] = True
        kwargs['bias'] = True
        kwargs['zero_init'] = True
        self.hidden_dim = hidden_dim
        super().__init__(*args, **kwargs)
    
    def init_weights_(self):
        """
        Initialize (W)eights and (b)iases
        """
        kwargs = {'dtype': self.dtype, 'device': self.device}
        self.layer0 = nn.Parameter(
            torch.zeros((self.num_heads, self.head_dim, self.hidden_dim), **kwargs)
        )
        self.layer1 = nn.Parameter(
            torch.zeros((self.num_heads, self.hidden_dim, self.feature_dim), **kwargs)
        )
        nn.init.kaiming_uniform_(self.layer0)
        nn.init.kaiming_uniform_(self.layer1)

        self.bias0 = nn.Parameter(torch.zeros((1, self.num_heads, 1, self.hidden_dim), **kwargs))
        self.bias1 = nn.Parameter(torch.zeros((1, self.num_heads, 1, self.feature_dim), **kwargs))
        nn.init.kaiming_uniform_(self.bias0)
        nn.init.kaiming_uniform_(self.bias1)

    def zero_init_with_skip_(self):
        with torch.no_grad():
            nn.init.zeros_(self.layer0)
            nn.init.zeros_(self.layer1)
            nn.init.zeros_(self.bias0)
            nn.init.zeros_(self.bias1)

    def zero_init_(self):
        assert NotImplementedError

    def forward(self, x: torch.Tensor):
        """
        Assume x.shape is (batch_size, num_heads, seq_len, head_dim)
        -> Down-project, apply nonlinearity, up-project; add skip connection
        """
        _x = torch.einsum('hde,bhld->bhle', self.layer0, x) + self.bias0
        _x = F.relu(_x)
        _x = torch.einsum('hef,bhle->bhlf', self.layer1, _x) + self.bias1
        return x + _x if self.skip_connection else _x



import torch
import thunderkittens as tk
from tqdm import tqdm

def generate_tensor(shape, mean, std, dtype, device):
    tensor = torch.randn(shape, dtype=dtype, device=device)
    magnitude = torch.norm(tensor, dim=-1, keepdim=True)
    scaled_tensor = tensor * (torch.randn(magnitude.shape, dtype=dtype, device=device) * std + mean) / magnitude
    return scaled_tensor.contiguous()

def tk_forward_test(Q, K, V, causal):
    O = torch.zeros_like(Q).contiguous()
    L = torch.zeros(Q.shape[0], Q.shape[1], Q.shape[2], 1, device=Q.device, dtype=torch.float)
    tk.mha_forward(Q, K, V, O, L, causal)
    return O, L

def check_consistency(b, h, n, d, causal, mean, std, num_iterations=100000):
    # Generate fixed inputs
    torch.manual_seed(0)
    Q = generate_tensor((b, h, n, d), mean, std, torch.bfloat16, 'cuda')
    K = generate_tensor((b, h, n, d), mean, std, torch.bfloat16, 'cuda')
    V = generate_tensor((b, h, n, d), mean, std, torch.bfloat16, 'cuda')

    # Initial run to get reference outputs
    ref_O, ref_L = tk_forward_test(Q, K, V, causal)

    max_diff_O = 0
    max_diff_L = 0

    for i in tqdm(range(num_iterations), desc="Checking consistency"):
        # Forward pass
        O, L = tk_forward_test(Q, K, V, causal)
        
        # Check differences
        current_diff_O = torch.abs(O - ref_O).max().item()
        current_diff_L = torch.abs(L - ref_L).max().item()
        
        max_diff_O = max(max_diff_O, current_diff_O)
        max_diff_L = max(max_diff_L, current_diff_L)
        
        if current_diff_O > 1e-7 or current_diff_L > 1e-7:
            print(f"Iteration {i}: Large difference detected")
            print(f"Current diff O: {current_diff_O}")
            print(f"Current diff L: {current_diff_L}")
            # breakpoint()

    return max_diff_O, max_diff_L

# Example usage
b, h, n, d = 1, 16, 768*8, 128
causal = True
mean = 1e-1
std = 1

max_diff_O, max_diff_L = check_consistency(b, h, n, d, causal, mean, std)

print(f"Maximum difference in O across iterations: {max_diff_O}")
print(f"Maximum difference in L across iterations: {max_diff_L}")


import torch
from tqdm import trange
import numpy as np
import sys
torch.set_grad_enabled(False)

N = 1024
B = 4
H = 1024
N1 = int(np.sqrt(N))
N_64 = 64

TESTNAME = sys.argv[1]

if TESTNAME in ['ones']:
    TESTNAME = f'ones_f{N}_H{H}_B{B}'
    u = (torch.ones((B, H, N), dtype=torch.bfloat16, device='cpu')).to(torch.float32) 
    k = (torch.ones((H, N), dtype=torch.bfloat16, device='cpu')).to(torch.float32)
    
elif TESTNAME in ['randn']:
    TESTNAME = f'randn_f{N}_H{H}_B{B}'
    torch.random.manual_seed(42)
    u = (torch.randn((B, H, N), dtype=torch.bfloat16, device='cpu')).to(torch.float32) 
    k = (torch.randn((H, N), dtype=torch.bfloat16, device='cpu')).to(torch.float32)

elif TESTNAME in ['arange']:
    TESTNAME = f'arange_f{N}_H{H}_B{B}'
    u = (torch.arange(B*H*N, dtype=torch.bfloat16, device='cpu')).to(torch.float32).reshape(B, H, N)
    k = (torch.arange(H*N, dtype=torch.bfloat16, device='cpu')).to(torch.float32).reshape(H, N)

else:
    print('Invalid test name')
    sys.exit(0)

def ref_fftconv(u, k, N):
    L = u.shape[-1]
    u_f = torch.fft.fft(u.float(), n = N)
    k_f = torch.fft.fft(k.float(), n = N)
    y_f = u_f * k_f
    y = torch.fft.ifft(y_f, n = N).real[..., :L].to(u.dtype).contiguous()
    return y

def fft_matrix(N):
    n = torch.arange(N)
    k = n.view(-1, 1)
    M = torch.exp(-2j * torch.pi * n * k / N)
    return M

def compute_twiddle_factors_fft(n, m):
    """Compute the twiddle factors of size n x m"""
    n_a = torch.arange(n).view(-1, 1)
    m_a = torch.arange(m)
    N = n * m
    M = torch.exp(-2j * torch.pi * n_a * m_a / N)
    return M

def ifft_matrix(N):
    n = torch.arange(N)
    k = n.view(-1, 1)
    M = torch.exp(2j * torch.pi * n * k / N)
    return M

def compute_twiddle_factors_ifft(n, m):
    """Compute the twiddle factors of size n x m"""
    n_a = torch.arange(n).view(-1, 1)
    m_a = torch.arange(m)
    N = n * m
    M = torch.exp(2j * torch.pi * n_a * m_a / N)
    return M


def our_reference_logic(u_reals, u_imags, k_fT_64x64_real, k_fT_64x64_imag, f_sqrt_64x64_real, f_sqrt_64x64_imag, f_inv_sqrt_64x64_real, f_inv_sqrt_64x64_imag, tw_real_64x64_real, tw_imag_64x64_imag, twinv_real_64x64_real, twinv_imag_64x64_imag):
    # breakpoint()
    B, H, N1, N1 = u_reals[0].shape

    # get the inputs
    x = torch.zeros(B, H, 64, 64).to(u_reals[0].device).to(torch.cfloat)
    x[:, :, :N1, :N1] = u_reals[0] + 1j * u_imags[0]
    x[:, :, :N1, N1:] = u_reals[1] + 1j * u_imags[1]
    x[:, :, N1:, :N1] = u_reals[2] + 1j * u_imags[2]
    x[:, :, N1:, N1:] = u_reals[3] + 1j * u_imags[3]

    f_sqrt_N_fft = torch.zeros(64, 64, device=f_sqrt_64x64_real.device).to(f_sqrt_64x64_real.dtype).to(torch.cfloat)
    f_sqrt_N_fft.real = f_sqrt_64x64_real
    f_sqrt_N_fft.imag = f_sqrt_64x64_imag

    f_sqrt_N_ifft = torch.zeros(64, 64, device=f_inv_sqrt_64x64_real.device).to(f_inv_sqrt_64x64_real.dtype).to(torch.cfloat)
    f_sqrt_N_ifft.real = f_inv_sqrt_64x64_real
    f_sqrt_N_ifft.imag = f_inv_sqrt_64x64_imag

    twiddle_factors_fft = torch.zeros(64, 64, device=tw_real_64x64_real.device).to(tw_real_64x64_real.dtype).to(torch.cfloat)
    twiddle_factors_fft.real = tw_real_64x64_real
    twiddle_factors_fft.imag = tw_imag_64x64_imag

    twiddle_factors_ifft = torch.zeros(64, 64, device=twinv_real_64x64_real.device).to(twinv_real_64x64_real.dtype).to(torch.cfloat)
    twiddle_factors_ifft.real = twinv_real_64x64_real
    twiddle_factors_ifft.imag = twinv_imag_64x64_imag

    k_f = torch.zeros(H, 64, 64, device=k_fT_64x64_real.device).to(k_fT_64x64_real.dtype).to(torch.cfloat)
    k_f.real = k_fT_64x64_real
    k_f.imag = k_fT_64x64_imag

    # run op
    x = x.reshape(B, H, 64, 64)
    x = f_sqrt_N_fft @ x
    
    x = x * twiddle_factors_fft
    x = x @ f_sqrt_N_fft

    k_f = k_f.reshape(H, 64, 64) 
    x = x * k_f

    x = x @ f_sqrt_N_ifft   
    x = x.transpose(-1, -2)
    x = x * twiddle_factors_ifft 

    x = x @ f_sqrt_N_ifft 
    x = x.transpose(-1, -2) 

    our_outputs = []
    out1 = x[:, :, :N1, :N1].reshape(B, H, N1, N1).real.to(torch.bfloat16).contiguous()
    out2 = x[:, :, :N1, N1:].reshape(B, H, N1, N1).real.to(torch.bfloat16).contiguous()
    out3 = x[:, :, N1:, :N1].reshape(B, H, N1, N1).real.to(torch.bfloat16).contiguous()
    out4 = x[:, :, N1:, N1:].reshape(B, H, N1, N1).real.to(torch.bfloat16).contiguous()
    our_outputs.append(out1)
    our_outputs.append(out2)
    our_outputs.append(out3)
    our_outputs.append(out4)
    return our_outputs


def pytorch_test(u, k, TESTNAME='all'):
    u_reals, u_imags = [], []
    u_real = u.to(torch.bfloat16)
    u_imag = torch.zeros_like(u, dtype=torch.bfloat16)
    u_reals.append(u_real)
    u_imags.append(u_imag)
    for i in range(3):
        u_new = torch.randn_like(u)*(i+2)
        u_real = u_new.to(torch.bfloat16)
        u_imag = torch.zeros_like(u, dtype=torch.bfloat16)
        u_reals.append(u_real)
        u_imags.append(u_imag)

    # fft matrix
    f_mat = fft_matrix(N1)
    f_real = f_mat.real.to(torch.bfloat16).contiguous()
    f_imag = f_mat.imag.to(torch.bfloat16).contiguous()

    f_sqrt_64x64_real = torch.zeros(64, 64, device=f_real.device).to(f_real.dtype)
    f_sqrt_64x64_real[:N1, :N1] = f_real
    f_sqrt_64x64_real[N1:, N1:] = f_real

    f_sqrt_64x64_imag = torch.zeros(64, 64, device=f_imag.device).to(f_imag.dtype)
    f_sqrt_64x64_imag[N1:, N1:] = f_imag
    f_sqrt_64x64_imag[:N1, :N1] = f_imag

    # ifft matrix
    finv_mat = ifft_matrix(N1)
    finv_real = finv_mat.real.to(torch.bfloat16).contiguous()
    finv_imag = finv_mat.imag.to(torch.bfloat16).contiguous()

    f_inv_sqrt_64x64_real = torch.zeros(64, 64, device=finv_real.device).to(finv_real.dtype)
    f_inv_sqrt_64x64_real[:N1, :N1] = finv_real
    f_inv_sqrt_64x64_real[N1:, N1:] = finv_real

    f_inv_sqrt_64x64_imag = torch.zeros(64, 64, device=finv_imag.device).to(finv_imag.dtype)
    f_inv_sqrt_64x64_imag[N1:, N1:] = finv_imag
    f_inv_sqrt_64x64_imag[:N1, :N1] = finv_imag

    # Normalization factor to make IFFT exact inverse of FFT
    tw = compute_twiddle_factors_fft(N1, N1) / N
    tw_real = tw.real.to(torch.bfloat16).contiguous()
    tw_imag = tw.imag.to(torch.bfloat16).contiguous()

    tw_real_64x64_real = torch.zeros(64, 64, device=tw_real.device).to(tw_real.dtype)
    tw_real_64x64_real[:N1, :N1] = tw_real
    tw_real_64x64_real[N1:, N1:] = tw_real
    tw_real_64x64_real[:N1, N1:] = tw_real
    tw_real_64x64_real[N1:, :N1] = tw_real

    tw_imag_64x64_imag = torch.zeros(64, 64, device=tw_imag.device).to(tw_imag.dtype)
    tw_imag_64x64_imag[:N1, :N1] = tw_imag
    tw_imag_64x64_imag[N1:, N1:] = tw_imag
    tw_imag_64x64_imag[:N1, N1:] = tw_imag
    tw_imag_64x64_imag[N1:, :N1] = tw_imag

    # twiddle inverse
    twinv = compute_twiddle_factors_ifft(N1, N1)
    twinv_real = twinv.real.to(torch.bfloat16).contiguous()
    twinv_imag = twinv.imag.to(torch.bfloat16).contiguous()

    twinv_real_64x64_real = torch.zeros(64, 64, device=twinv_real.device).to(twinv_real.dtype)
    twinv_real_64x64_real[:N1, :N1] = twinv_real
    twinv_real_64x64_real[N1:, N1:] = twinv_real
    twinv_real_64x64_real[:N1, N1:] = twinv_real
    twinv_real_64x64_real[N1:, :N1] = twinv_real

    twinv_imag_64x64_imag = torch.zeros(64, 64, device=twinv_imag.device).to(twinv_imag.dtype)
    twinv_imag_64x64_imag[:N1, :N1] = twinv_imag
    twinv_imag_64x64_imag[N1:, N1:] = twinv_imag
    twinv_imag_64x64_imag[:N1, N1:] = twinv_imag
    twinv_imag_64x64_imag[N1:, :N1] = twinv_imag

    # Compute the regular FFT if the seq len isn't 512 or 2048
    k_f = torch.fft.fft(k.float(), n = N)
    k_fT = k_f.reshape(H, N1, N1).transpose(-1, -2)
    kfT_real = k_fT.real.to(torch.bfloat16).contiguous()
    kfT_imag = k_fT.imag.to(torch.bfloat16).contiguous()

    k_fT_64x64_real = torch.zeros(H, 64, 64, device=kfT_real.device).to(torch.bfloat16)
    k_fT_64x64_real[:, :N1, :N1] = kfT_real.to(k_fT_64x64_real.dtype)
    k_fT_64x64_real[:, N1:, N1:] = kfT_real.to(k_fT_64x64_real.dtype)
    k_fT_64x64_real[:, :N1, N1:] = kfT_real.to(k_fT_64x64_real.dtype)
    k_fT_64x64_real[:, N1:, :N1] = kfT_real.to(k_fT_64x64_real.dtype)

    k_fT_64x64_imag = torch.zeros(H, 64, 64, device=kfT_imag.device).to(torch.bfloat16)
    k_fT_64x64_imag[:, :N1, :N1] = kfT_imag.to(k_fT_64x64_imag.dtype)
    k_fT_64x64_imag[:, N1:, N1:] = kfT_imag.to(k_fT_64x64_imag.dtype)
    k_fT_64x64_imag[:, :N1, N1:] = kfT_imag.to(k_fT_64x64_imag.dtype)
    k_fT_64x64_imag[:, N1:, :N1] = kfT_imag.to(k_fT_64x64_imag.dtype)

    o_reals = []
    for i in range(4):
        o_real = ref_fftconv(u_reals[i], k, N)
        o_real = o_real.reshape(B, H, N1, N1).to(torch.bfloat16).contiguous()
        o_reals.append(o_real)

    u_reals_reshaped, u_imags_reshaped = [], []
    for i in range(4):
        u_reals_reshaped.append(u_reals[i].reshape(B, H, N1, N1).to(torch.bfloat16).contiguous())
        u_imags_reshaped.append(u_imags[i].reshape(B, H, N1, N1).to(torch.bfloat16).contiguous())
    
    return u_reals_reshaped, u_imags_reshaped, k_fT_64x64_real, k_fT_64x64_imag, f_sqrt_64x64_real, f_sqrt_64x64_imag, f_inv_sqrt_64x64_real, f_inv_sqrt_64x64_imag, tw_real_64x64_real, tw_imag_64x64_imag, twinv_real_64x64_real, twinv_imag_64x64_imag, o_reals


def check_our_outputs(our_outputs, o_reals):
    for i in range(4):
        all_close = torch.allclose(our_outputs[i], o_reals[i], atol=1e-3, rtol=1e-3)
        max_diff = torch.max(torch.abs(our_outputs[i] - o_reals[i]))
        mean_diff = torch.mean(torch.abs(our_outputs[i] - o_reals[i]))
        avg_value = torch.mean(torch.abs(o_reals[i]))
        print(f'Output {i+1} all close: {all_close}, max diff: {max_diff}, mean diff: {mean_diff}, avg value: {avg_value}')

u_reals_reshaped, u_imags_reshaped, k_fT_64x64_real, k_fT_64x64_imag, f_sqrt_64x64_real, f_sqrt_64x64_imag, f_inv_sqrt_64x64_real, f_inv_sqrt_64x64_imag, tw_real_64x64_real, tw_imag_64x64_imag, twinv_real_64x64_real, twinv_imag_64x64_imag, o_reals = pytorch_test(u, k, TESTNAME)

# run a check with these values
our_outputs = our_reference_logic(u_reals_reshaped, u_imags_reshaped, k_fT_64x64_real, k_fT_64x64_imag, f_sqrt_64x64_real, f_sqrt_64x64_imag, f_inv_sqrt_64x64_real, f_inv_sqrt_64x64_imag, tw_real_64x64_real, tw_imag_64x64_imag, twinv_real_64x64_real, twinv_imag_64x64_imag)
check_our_outputs(our_outputs, o_reals)


with open(f'{TESTNAME}.txt', 'w') as f:
    # u reals 
    for i in range(len(u_reals_reshaped)):
        u_real_f = u_reals_reshaped[i].to(torch.float32).flatten().cpu().numpy()
        for j in trange(u_real_f.shape[0]):
            f.write(repr(u_real_f[j]))
            f.write(' ')
    for i in range(len(u_reals_reshaped)):
        u_imag_f = u_imags_reshaped[i].to(torch.float32).flatten().cpu().numpy()
        for j in trange(u_imag_f.shape[0]):
            f.write(repr(u_imag_f[j]))
            f.write(' ')

    kfT_real_f = k_fT_64x64_real.to(torch.float32).flatten().cpu().numpy()
    kfT_imag_f = k_fT_64x64_imag.to(torch.float32).flatten().cpu().numpy()
    for i in trange(kfT_real_f.shape[0]):
        f.write(repr(kfT_real_f[i]))
        f.write(' ')
    for i in trange(kfT_imag_f.shape[0]):
        f.write(repr(kfT_imag_f[i]))
        f.write(' ')

    f_real_f = f_sqrt_64x64_real.to(torch.float32).flatten().cpu().numpy()
    f_imag_f = f_sqrt_64x64_imag.to(torch.float32).flatten().cpu().numpy()
    for i in trange(f_real_f.shape[0]):
        f.write(repr(f_real_f[i]))
        f.write(' ')
    for i in trange(f_imag_f.shape[0]):
        f.write(repr(f_imag_f[i]))
        f.write(' ')


    finv_real_f = f_inv_sqrt_64x64_real.to(torch.float32).flatten().cpu().numpy()
    finv_imag_f = f_inv_sqrt_64x64_imag.to(torch.float32).flatten().cpu().numpy()
    for i in trange(finv_real_f.shape[0]):
        f.write(repr(finv_real_f[i]))
        f.write(' ')
    for i in trange(finv_imag_f.shape[0]):
        f.write(repr(finv_imag_f[i]))
        f.write(' ')

    tw_real_f = tw_real_64x64_real.to(torch.float32).flatten().cpu().numpy()
    tw_imag_f = tw_imag_64x64_imag.to(torch.float32).flatten().cpu().numpy()
    for i in trange(tw_real_f.shape[0]):
        f.write(repr(tw_real_f[i]))
        f.write(' ')
    for i in trange(tw_imag_f.shape[0]):
        f.write(repr(tw_imag_f[i]))
        f.write(' ')
    
    twinv_real_f = twinv_real_64x64_real.to(torch.float32).flatten().cpu().numpy()
    twinv_imag_f = twinv_imag_64x64_imag.to(torch.float32).flatten().cpu().numpy()
    for i in trange(twinv_real_f.shape[0]):
        f.write(repr(twinv_real_f[i]))
        f.write(' ')
    for i in trange(twinv_imag_f.shape[0]):
        f.write(repr(twinv_imag_f[i]))
        f.write(' ')

    for i in range(len(o_reals)):
        o_real = o_reals[i]
        o_real_f = o_real.to(torch.float32).flatten().cpu().numpy()
        for j in trange(o_real_f.shape[0]):
            f.write(repr(o_real_f[j]))
            f.write(' ')
    


import example_bind
import torch

input = torch.ones(16, 1024, 32, 64, device='cuda')
output = torch.zeros_like(input)
example_bind.copy_kernel(input, output)
print(output.mean(), '\n')

output = torch.zeros_like(input)
example_bind.wrapped_copy_kernel(input, output)
print(output.mean())


"""
LoLCATs attention combining sliding window and linear attentions
- Using the TK "terracing" arrangement
- Training over long sequences with fixed memory with recurrent view
- During attention transfer, use Flash Attention to compute softmax attention outputs

For each layer: 
- We first compute (softmax) attention over sliding windows
- We then compute standard linear attention to "fill in" the earlier parts
- We combine to model the entire sequence
"""
from typing import Optional, Tuple
import torch
import torch.nn.functional as F

from transformers.cache_utils import Cache
try:
    from transformers.modeling_flash_attention_utils import _flash_attention_forward
except ModuleNotFoundError:
    _flash_attention_forward = None  # Transformers v4.36

from src.model.rotary import apply_rotary_pos_emb
from .linear_window_attention_tk import LolcatsTKWindowAttention
from .linear_attention import softmax_attention


class LolcatsTKWindowLongAttention(LolcatsTKWindowAttention):
    """
    Lolcats attention combining sliding window and linear attention
    """
    def __init__(self, remove_base_attn=True, **kwargs):
        # keep self.base_attn for Flash Attention inference
        super().__init__(remove_base_attn=True, **kwargs)

    def forward(self,
                hidden_states: torch.Tensor,
                attention_mask: Optional[torch.Tensor] = None,
                position_ids: Optional[torch.LongTensor] = None,
                past_key_value: Optional[Cache] = None,
                output_attentions: bool = False,
                use_cache: bool = False,
                **kwargs,
               ) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[Tuple[torch.Tensor]]]:
        """
        Forward pass with the option to compute attention weights multiple ways
        if self.train_attention is True
        -> Consistent with HuggingFace Transformers for easy use with their pretrained models
        """
        b, l, _ = hidden_states.size()
        if self.train_attention and self.base_inference:
            with torch.no_grad():
                # print(hidden_states.shape)
                _y_true = flash_attention_2(self,  # self.base_attn,
                                            hidden_states=hidden_states,
                                            attention_mask=None,
                                            position_ids=position_ids,
                                            past_key_value=None,
                                            output_attentions=False,
                                            # output_hidden_states=False,
                                            use_cache=False)[0]
                # _y_true.shape is (batch_size, seq_len, num_heads, head_dim)
                y_true = _y_true.reshape(b, l, -1).contiguous()
                y_true = self.o_proj(y_true)
                layer_io = (hidden_states, _y_true)  # hack
                # layer_io = (hidden_states.cpu(), _y_true.cpu())  # hack
                return y_true, layer_io, None

        q, k, v, kv_seq_len = self.process_qkv(hidden_states, attention_mask,
                                               position_ids, past_key_value)
        f_q, f_k = self.feature_map_q(q), self.feature_map_k(k)

        # attention_mask = None  # For now this is always True
        if past_key_value is None:  # Regular training
            window_factors = F.sigmoid(self.window_factors)
            linear_factors = 1 - window_factors if self.affine_attention_factors else 1
            y_pred, a_pred = self.quadratic_attention(q, k, f_q, f_k, v, 
                                                      window_factors, linear_factors,
                                                      window_size=self.window_size,)
        else:
            past_key_value.window_size = self.decode_window_size
            if f_q.shape[2] == 1 and kv_seq_len > 1 and not self.training:  # Generating
                assert use_cache is True
                _kv = past_key_value.update_for_decoding(k, v, self.layer_idx,
                                                         self.feature_map_k,
                                                         dtype=q.dtype)
                k_cache, v_cache, f_kv_state, f_k_state = _kv

                # Sliding window + linear attention decode
                window_factors = F.sigmoid(self.window_factors)
                linear_factors = 1 - window_factors if self.affine_attention_factors else 1

                a_sm = torch.einsum('bhmd,bhnd->bhmn', q.float(), k_cache.float()) * (k.shape[-1] ** -0.5)
                # a_sm = torch.softmax(a_sm, dim=-1)
                a_sm_max = torch.amax(a_sm, dim=-1, keepdim=True)
                a_sm   = window_factors * torch.exp(a_sm - a_sm_max)
                sum_sm = a_sm.sum(dim=-1, keepdim=True)

                y_pred = (torch.einsum('bhmn,bhnd->bhmd', a_sm, v_cache.float())
                          + linear_factors * torch.einsum('bhlf,bhfd->bhld', f_q.float(), f_kv_state.float()))
                sum_ln = linear_factors * torch.einsum('bhlf,bhnf->bhl', f_q.float(), f_k_state.float())[..., None]
                y_pred = (y_pred / (sum_sm + sum_ln)).to(q.dtype) 

            else:  # Stateful training
                if self.state_grad_enabled and self.layer_idx == 0:
                    print(f'\n position_ids: [{position_ids[0, 0]}, {position_ids[0, -1]}]')
                    print(f'q.shape: {q.shape}, k.shape: {k.shape}, v.shape: {v.shape}')
                try:
                    kv_state = past_key_value.kv_states[self.layer_idx]
                    k_state  = past_key_value.k_states[self.layer_idx]
                except IndexError:
                    kv_state, k_state = None, None
                window_factors = F.sigmoid(self.window_factors)
                linear_factors = 1 - window_factors if self.affine_attention_factors else 1
                y_pred, a_pred = self.quadratic_attention(q, k, f_q, f_k, v,
                                                          window_factors, linear_factors,
                                                          window_size=self.window_size,
                                                          kv_state=kv_state,
                                                          k_state=k_state,)
                # Save and update KV cache and states
                # past_key_value.update(k, v.detach(), self.layer_idx,
                #                       fmap_key_states=f_k.detach(),
                #                       accumulate_in_fp32=True)
                past_key_value.update(k, v, self.layer_idx,
                                      fmap_key_states=f_k,
                                      accumulate_in_fp32=True)

        # Concatenate heads and apply output projection
        _y_pred = y_pred.transpose(1, 2).contiguous()
        y_pred = self.o_proj(_y_pred.view(b, l, self.hidden_size))

        if self.train_attention:
            with torch.no_grad():
                a_true = softmax_attention(q, k, None, causal=True)[1]     
            attn_weights = (_y_pred, (a_pred, a_true))
        else:
            attn_weights = _y_pred  # flash_attn outputs are shape (b, l, h, d)
        return y_pred, attn_weights, past_key_value


# -----------------
# Flash Attention 2
# -----------------

def flash_attention_2(self,
                      hidden_states: torch.Tensor,
                      attention_mask: Optional[torch.LongTensor] = None,
                      position_ids: Optional[torch.LongTensor] = None,
                      past_key_value: Optional[Cache] = None,
                      output_attentions: bool = False,
                      use_cache: bool = False,
                      cache_position: Optional[torch.LongTensor] = None,
                     ) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[Tuple[torch.Tensor]]]:
    """
    Wrapper for LlamaFlashAttention2
    Copied and modified from HF Transformers v4.36 and v4.43 implementations
    - (4.43) https://github.com/huggingface/transformers/blob/868d36d29ec132deeaaf8571b25b6a1b911d0145/src/transformers/models/llama/modeling_llama.py#L402
    - (4.36) https://github.com/huggingface/transformers/blob/a7cab3c283312b8d4de5df3bbe719971e24f4281/src/transformers/models/llama/modeling_llama.py#L456
    """
    output_attentions = False

    bsz, q_len, _ = hidden_states.size()

    query_states = self.q_proj(hidden_states)
    key_states = self.k_proj(hidden_states)
    value_states = self.v_proj(hidden_states)

    # Flash attention requires the input to have the shape
    # batch_size x seq_length x head_dim x hidden_dim
    # therefore we just need to keep the original shape
    query_states = query_states.view(bsz, q_len, self.num_heads, self.head_dim).transpose(1, 2)
    key_states = key_states.view(bsz, q_len, self.num_key_value_heads, self.head_dim).transpose(1, 2)
    value_states = value_states.view(bsz, q_len, self.num_key_value_heads, self.head_dim).transpose(1, 2)

    try: # As in Transformers v4.36
        kv_seq_len = key_states.shape[-2]
        if past_key_value is not None:
            kv_seq_len += past_key_value.get_usable_length(kv_seq_len, self.layer_idx)
        cos, sin = self.rotary_emb(key_states, seq_len=kv_seq_len)    
        query_states, key_states = apply_rotary_pos_emb(query_states, key_states, cos, sin, position_ids)
    except:  # As in Transformers v4.39
        cos, sin = self.rotary_emb(key_states, position_ids)
        query_states, key_states = apply_rotary_pos_emb(query_states, key_states, cos, sin)

    if past_key_value is not None:
        # sin and cos are specific to RoPE models; cache_position needed for the static cache
        cache_kwargs = {"sin": sin, "cos": cos, "cache_position": cache_position}
        key_states, value_states = past_key_value.update(key_states, value_states, self.layer_idx, cache_kwargs)

    # TODO: These transpose are quite inefficient but Flash Attention requires the layout [batch_size, sequence_length, num_heads, head_dim]. We would need to refactor the KV cache
    # to be able to avoid many of these transpose/reshape/view.
    query_states = query_states.transpose(1, 2)
    key_states = key_states.transpose(1, 2)
    value_states = value_states.transpose(1, 2)

    dropout_rate = self.attention_dropout if self.training else 0.0

    # In PEFT, usually we cast the layer norms in float32 for training stability reasons
    # therefore the input hidden states gets silently casted in float32. Hence, we need
    # cast them back in the correct dtype just to be sure everything works as expected.
    # This might slowdown training & inference so it is recommended to not cast the LayerNorms
    # in fp32. (LlamaRMSNorm handles it correctly)

    input_dtype = query_states.dtype
    if input_dtype == torch.float32:
        if torch.is_autocast_enabled():
            target_dtype = torch.get_autocast_gpu_dtype()
        # Handle the case where the model is quantized
        elif hasattr(self.config, "_pre_quantization_dtype"):
            target_dtype = self.config._pre_quantization_dtype
        else:
            target_dtype = self.q_proj.weight.dtype

        logger.warning_once(
            f"The input hidden states seems to be silently casted in float32, this might be related to"
            f" the fact you have upcasted embedding or layer norm layers in float32. We will cast back the input in"
            f" {target_dtype}."
        )

        query_states = query_states.to(target_dtype)
        key_states = key_states.to(target_dtype)
        value_states = value_states.to(target_dtype)

    if getattr(self, '_flash_attention_forward', False):
        attn_output = self._flash_attention_forward(
            query_states, key_states, value_states, attention_mask, q_len, dropout=dropout_rate,
            is_causal=True,
        )
    else:
        attn_output = _flash_attention_forward(
            query_states,
            key_states,
            value_states,
            attention_mask,
            q_len,
            dropout=0, # dropout_rate,
            sliding_window=getattr(self, "sliding_window", None),
            use_top_left_mask=self._flash_attn_uses_top_left_mask,
            is_causal=True,
        )
    return attn_output, past_key_value


### ADD TO THIS TO REGISTER NEW KERNELS
sources = {
    'attn': {
        'source_files': {
            'h100': 'kernels/attn/h100/h100.cu' # define these source files for each GPU target desired.
        }
    },
    'hedgehog': {
        'source_files': {
            'h100': 'kernels/hedgehog/hh.cu'
        }
    },
    'based': {
        'source_files': {
            'h100': [
                'kernels/based/lin_attn_h100.cu',
            ],
            '4090': [
                'kernels/based/lin_attn_4090.cu',
            ]
        }
    },
    'cylon': {
        'source_files': {
            'h100': 'kernels/cylon/cylon.cu'
        }
    },
    'flux': {
        'source_files': {
            'h100': [
                'kernels/flux/flux_gate.cu',
                'kernels/flux/flux_gelu.cu'
            ]
        }
    },
    'fftconv': {
        'source_files': {
            'h100': 'kernels/fftconv/pc/pc.cu'
        }
    },
    'fused_rotary': {
        'source_files': {
            'h100': 'kernels/rotary/pc.cu'
        }
    },
    'fused_layernorm': {
        'source_files': {
            'h100': 'kernels/layernorm/non_pc/layer_norm.cu'
        }
    },
    'mamba2': {
        'source_files': {
            'h100': 'kernels/mamba2/pc.cu'
        }
    },
    'fp8_gemm': {
        'source_files': {
            'h100': 'kernels/matmul/FP8/matmul.cu'
        }
    },
}

### WHICH KERNELS DO WE WANT TO BUILD?
# (oftentimes during development work you don't need to redefine them all.)
# kernels = ['attn', 'mamba2', 'hedgehog', 'fftconv', 'fused_rotary', 'based', 'fused_layernorm']
kernels = ['fp8_gemm']

### WHICH GPU TARGET DO WE WANT TO BUILD FOR?
target = 'h100'



"""
Quick demo of linearized LLM generations
"""
from typing import Optional, List
import time
import argparse
import torch

from omegaconf import OmegaConf

from transformers import TextStreamer, TextIteratorStreamer, AutoTokenizer

from src.utils.setup import seed_everything
from src.utils.logging import print_header
from src.model.pretrained import get_pretrained_loader


system_prompt = """{prompt}"""


def get_args():
    parser = argparse.ArgumentParser()
    # Model load + setup
    parser.add_argument("--model_config_path", type=str)
    parser.add_argument("--config_dir", type=str, default='configs')
    parser.add_argument("--seed", type=int, default=42)

    # Inference
    parser.add_argument("--model_type", type=str, default='flash_attention_2')

    # Generation
    parser.add_argument("--num_generations", type=int, default=1)
    parser.add_argument("--top_k", type=int, default=1.0)
    parser.add_argument("--top_p", type=float, default=1.0)
    parser.add_argument("--max_new_tokens", type=int, default=2) 

    # Miscellaneous
    parser.add_argument("--benchmark", action='store_true', default=False)
    parser.add_argument("--print_model", action='store_true', default=False)
    parser.add_argument("--debug", action='store_true', default=False)
    parser.add_argument("--huggingface_token", type=str, default=None)

    args = parser.parse_args()
    return args


class BatchTextIteratorStreamer(TextIteratorStreamer):
    """
    Copied from https://discuss.huggingface.co/t/textiteratorstreamer-compatibility-with-batch-processing/46763/2
    """
    def __init__(self, 
                 tokenizer: AutoTokenizer, 
                 batch_size: int, 
                 skip_prompt: bool = False, 
                 timeout: Optional[float] = None, 
                 **decode_kwargs: any):
        super().__init__(tokenizer, skip_prompt, timeout, **decode_kwargs)
        self.batch_size = batch_size
        self.token_cache = [[] for _ in range(batch_size)]
        self.print_len = [0 for _ in range(batch_size)]
        self.generate_exception = None
        self.go_up = 0 + batch_size
        self.stop_signal = tokenizer.eos_token

    def put(self, value):
        if len(value.shape) != 2:
            value = torch.reshape(value, (self.batch_size, value.shape[0] // self.batch_size))

        if self.skip_prompt and self.next_tokens_are_prompt:
            self.next_tokens_are_prompt = False
            return

        printable_texts = list()
        for idx in range(self.batch_size):
            self.token_cache[idx].extend(value[idx].tolist())
            text = self.tokenizer.decode(self.token_cache[idx], **self.decode_kwargs)

            if text.endswith("\n"):
                printable_text = text[self.print_len[idx] :]
                self.token_cache[idx] = []
                self.print_len[idx] = 0
                self.go_up += 1
                # If the last token is a CJK character, we print the characters.
            elif len(text) > 0 and self._is_chinese_char(ord(text[-1])):
                printable_text = text[self.print_len[idx] :]
                self.print_len[idx] += len(printable_text)
            else:
                printable_text = text[self.print_len[idx] : text.rfind(" ") + 1]
                self.print_len[idx] += len(printable_text)
            printable_texts.append(printable_text)

        self.on_finalized_text(printable_texts)

    def end(self):
        printable_texts = list()
        for idx in range(self.batch_size):
            if len(self.token_cache[idx]) > 0:
                text = self.tokenizer.decode(self.token_cache[idx], **self.decode_kwargs)
                printable_text = text[self.print_len[idx] :]
                self.token_cache[idx] = []
                self.print_len[idx] = 0
            else:
                printable_text = ""
            printable_texts.append(printable_text)

        self.next_tokens_are_prompt = True
        self.on_finalized_text(printable_texts, stream_end=True)

    def on_finalized_text(self, texts: List[str], stream_end: bool = False):
        self.text_queue.put(texts, timeout=self.timeout)
        if stream_end:
            self.text_queue.put(self.stop_signal, timeout=self.timeout)

        try:
            text = [
                ''.join([x[i] if i < len(x) else self.stop_signal 
                         for x in self.text_queue.queue ]) 
                for i in range(len(self.text_queue.queue[0]))
            ]
            text = '\n------------\n'.join(text)
            go_up = "\033[F" * self.go_up  # len(text)  # Goes up this many lines
            print(f'{text}', flush=True, end="" if not stream_end else None)

        except Exception as e:
            print(self.stop_signal)

def count_params(module) -> int:
    return sum(p.numel() for p in module.parameters())

def load_model_from_checkpoint(model_type: str,
                               model_config_path: str = None,
                               huggingface_token: str = None):

    print(f"Using model type: {model_type}")

    model_config = OmegaConf.load(model_config_path)

    # Load initial transformer model
    model_config.model.pretrained_model_name_or_path = "Qwen/Qwen2.5-7B-Instruct"
    # breakpoint()
    model_loader = get_pretrained_loader(**model_config.model, 
                                         huggingface_token=huggingface_token)
    tokenizer = model_loader.load_tokenizer()
    tokenizer.pad_token_id = tokenizer.eos_token_id
    tokenizer.padding_side = 'left'
    model = model_loader.load(model_type = model_type)
    return model, model_config, tokenizer

def main():
    args = get_args()
    seed_everything(args.seed)
    model, model_config, tokenizer = load_model_from_checkpoint(
        args.model_type,
        args.model_config_path, 
        args.huggingface_token,
    )
    model = model.to('cuda')
    model.eval()
    input_len = len(tokenizer(system_prompt)['input_ids'])

    while True:
        print(f'\n>> Generating {args.num_generations} responses in parallel')
        prompt = input(f'>> Your prompt: (or cmd-c to quit)... ')

        # prompt hard coded to 192; no padding.
        # prompt = "Create a summary of the following passage: London is the capital city of England and the United Kingdom. It is a leading global city with strengths in the arts, commerce, education, entertainment, fashion, finance, healthcare, media, professional services, research and development, tourism, and transport all contributing to its prominence. It is one of the most populous cities in the world, with an estimated population of 8.9 million in 2019. At its centre stand the imposing Houses of Parliament, the iconic ‘Big Ben’ clock tower and Westminster Abbey, site of British monarch coronations. Across the Thames River, the London Eye observation wheel provides panoramic views of the South Bank cultural complex, and the entire city. London exerts a strong influence on world art,"
        #  entertainment, fashion, commerce, finance, education, healthcare, media, science, technology, tourism, transport, and communications. London's cultures encompass over 300 languages. The last time"

        all_prompts = [system_prompt.format(prompt=prompt)] * args.num_generations

        if args.num_generations == 1:
            streamer = TextStreamer(tokenizer, skip_prompt=True,
                                    decode_kwargs={'skip_special_tokens': True})
        else:
            streamer = BatchTextIteratorStreamer(tokenizer=tokenizer, 
                                                 batch_size=args.num_generations,
                                                 skip_prompt=True,)
    
        with torch.no_grad():
            model_input = tokenizer(all_prompts, return_tensors="pt").to(model.device)
            print(model_input['input_ids'].shape)

            if args.benchmark:
                torch.cuda.synchronize()
                start_time = time.time() 
            model_output = model.generate(**model_input, use_cache=True, 
                                          max_new_tokens=args.max_new_tokens, 
                                          do_sample=False,
                                          top_k=args.top_k,
                                          top_p=args.top_p,
                                          num_return_sequences=1,
                                          pad_token_id=tokenizer.eos_token_id,
                                          streamer=streamer)
            if args.benchmark:
                torch.cuda.synchronize()
                elapsed = time.time() - start_time
                total_tokens = (model_output != tokenizer.eos_token_id).sum().item()
                print_header('(Coarse) stats for nerds')
                print(f'├── Model data type:                      {model.dtype}')
                print(f'├── Time of longest response:             {elapsed:.3f} sec')
                print(f'├── Total tokens processed + generated:   {total_tokens}')
                print(f'├── Throughput (lagged by last response): {total_tokens / elapsed:.3f} tokens/sec')
                
        if args.benchmark:
            break

if __name__ == '__main__':
    main()





# coding=utf-8
# Copyright 2022 EleutherAI and the HuggingFace Inc. team. All rights reserved.
#
# This code is based on EleutherAI's GPT-NeoX library and the GPT-NeoX
# and OPT implementations in this library. It has been modified from its
# original forms to accommodate minor architectural differences compared
# to GPT-NeoX and OPT used by the Meta AI team that trained the model.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Rotary embeddings. Same as usual for Transformer models.

Note these are modified from HF Transformers v4.36, from:
- transformers/models/llama/modeling_llama.py or transformers/models/mistral/modeling_mistral.py
- i.e., https://github.com/huggingface/transformers/blob/a7cab3c283312b8d4de5df3bbe719971e24f4281/src/transformers/models/llama/modeling_llama.py#L123
"""
import torch
import torch.nn as nn


def get_rotary_embeddings(rope_scaling_type: str = None,
                          head_dim: int = 128,
                          max_position_embeddings: int = 4096,
                          rope_theta: float = 10000.0,
                          rope_scaling_factor: float = 1.0,
                          device: torch.device = None,
                         ) -> nn.Module:
    """Return rotary embedding object"""
    if rope_scaling_type is None:
        return RotaryEmbedding(
            head_dim,
            max_position_embeddings=max_position_embeddings,
            base=rope_theta,
            device=device,
        )
    elif rope_scaling_type == "linear":
        return LinearScalingRotaryEmbedding(
            head_dim,
            max_position_embeddings=max_position_embeddings,
            scaling_factor=rope_scaling_factor,
            base=rope_theta,
            device=device,
        )
    elif rope_scaling_type == "dynamic":
        return DynamicNTKScalingRotaryEmbedding(
            head_dim,
            max_position_embeddings=max_position_embeddings,
            scaling_factor=rope_scaling_factor,
            base=rope_theta,
            device=device,
        )
    else:
        raise NotImplementedError(f'Sorry rope_scaling_type == "{rope_scaling_type}" not implemented.')


# Copied from transformers.models.mistral.modeling_mistral (llama.modeling_llama at v4.36)
def rotate_half(x):
    """Rotates half the hidden dims of the input."""
    x1 = x[..., : x.shape[-1] // 2]
    x2 = x[..., x.shape[-1] // 2 :]
    return torch.cat((-x2, x1), dim=-1)


# Copied from transformers.models.mistral.modeling_mistral (llama.modeling_llama at v4.36)
def apply_rotary_pos_emb(q, k, cos, sin, position_ids=None, unsqueeze_dim=1):
    """Applies Rotary Position Embedding to the query and key tensors."""
    if position_ids is not None:
        cos, sin = cos[position_ids], sin[position_ids]
    cos = cos.unsqueeze(unsqueeze_dim)
    sin = sin.unsqueeze(unsqueeze_dim)
    q_embed = (q * cos) + (rotate_half(q) * sin)
    k_embed = (k * cos) + (rotate_half(k) * sin)
    return q_embed, k_embed


# Modified from transformers.models.mistral.modeling_mistral (llama.modeling_llama at v4.36)
class RotaryEmbedding(nn.Module):
    """Original Rotary Embeddings from RoFormer https://arxiv.org/abs/2104.09864"""
    def __init__(self, dim, max_position_embeddings=2048, base=10000, device=None):
        super().__init__()

        self.dim = dim
        self.max_position_embeddings = max_position_embeddings
        self.base = base
        inv_freq = 1.0 / (self.base ** (torch.arange(0, self.dim, 2).float().to(device) / self.dim))
        self.register_buffer("inv_freq", inv_freq, persistent=False)

        # Build here to make `torch.jit.trace` work.
        self._set_cos_sin_cache(
            seq_len=max_position_embeddings, device=self.inv_freq.device, dtype=torch.get_default_dtype()
        )

    def _set_cos_sin_cache(self, seq_len, device, dtype):
        self.max_seq_len_cached = seq_len
        t = torch.arange(self.max_seq_len_cached, device=device, dtype=self.inv_freq.dtype)

        freqs = torch.outer(t, self.inv_freq)
        # Different from paper, but it uses a different permutation in order to obtain the same calculation
        emb = torch.cat((freqs, freqs), dim=-1)
        self.register_buffer("cos_cached", emb.cos().to(dtype), persistent=False)
        self.register_buffer("sin_cached", emb.sin().to(dtype), persistent=False)

    def forward(self, x, seq_len=None):
        """
        Compute rotary embeddings
        """
        # x: [bs, num_attention_heads, seq_len, head_size]
        if seq_len > self.max_seq_len_cached:
            self._set_cos_sin_cache(seq_len=seq_len, device=x.device, dtype=x.dtype)

        return (
            self.cos_cached[:seq_len].to(dtype=x.dtype),
            self.sin_cached[:seq_len].to(dtype=x.dtype),
        )


# Copied from transformers/models/llama/modeling_llama.py at v4.36
class LinearScalingRotaryEmbedding(RotaryEmbedding):
    """RotaryEmbedding extended with linear scaling. Credits to the Reddit user /u/kaiokendev"""

    def __init__(self, dim, max_position_embeddings=2048, base=10000, device=None, scaling_factor=1.0):
        self.scaling_factor = scaling_factor
        super().__init__(dim, max_position_embeddings, base, device)

    def _set_cos_sin_cache(self, seq_len, device, dtype):
        self.max_seq_len_cached = seq_len
        t = torch.arange(self.max_seq_len_cached, device=device, dtype=self.inv_freq.dtype)
        t = t / self.scaling_factor

        freqs = torch.outer(t, self.inv_freq)
        # Different from paper, but it uses a different permutation in order to obtain the same calculation
        emb = torch.cat((freqs, freqs), dim=-1)
        self.register_buffer("cos_cached", emb.cos().to(dtype), persistent=False)
        self.register_buffer("sin_cached", emb.sin().to(dtype), persistent=False)


# Copied from transformers/models/llama/modeling_llama.py at v4.36
class DynamicNTKScalingRotaryEmbedding(RotaryEmbedding):
    """RotaryEmbedding extended with Dynamic NTK scaling. Credits to the Reddit users /u/bloc97 and /u/emozilla"""

    def __init__(self, dim, max_position_embeddings=2048, base=10000, device=None, scaling_factor=1.0):
        self.scaling_factor = scaling_factor
        super().__init__(dim, max_position_embeddings, base, device)

    def _set_cos_sin_cache(self, seq_len, device, dtype):
        self.max_seq_len_cached = seq_len

        if seq_len > self.max_position_embeddings:
            base = self.base * (
                (self.scaling_factor * seq_len / self.max_position_embeddings) - (self.scaling_factor - 1)
            ) ** (self.dim / (self.dim - 2))
            inv_freq = 1.0 / (base ** (torch.arange(0, self.dim, 2).float().to(device) / self.dim))
            self.register_buffer("inv_freq", inv_freq, persistent=False)

        t = torch.arange(self.max_seq_len_cached, device=device, dtype=self.inv_freq.dtype)

        freqs = torch.outer(t, self.inv_freq)
        # Different from paper, but it uses a different permutation in order to obtain the same calculation
        emb = torch.cat((freqs, freqs), dim=-1)
        self.register_buffer("cos_cached", emb.cos().to(dtype), persistent=False)
        self.register_buffer("sin_cached", emb.sin().to(dtype), persistent=False)




from thunderkittens import tk_flux_linear_gate, tk_flux_linear_gelu
import torch
import time

from triton.testing import do_bench

repeats = 100
warmups = 5

def ref_gate(x, linear, gate, y):
    return linear(x) * gate + y

def tk_gate(x, linear, gate, y):
    return tk_flux_linear_gate(x, linear.weight, linear.bias, gate, y)

def ref_gelu(x, linear):
    return torch.nn.functional.gelu(linear(x), approximate='tanh')

def tk_gelu(x, linear):
    return tk_flux_linear_gelu(x, linear.weight, linear.bias)

def benchmark_gated(M, N, K):
    with torch.device('cuda'):
        x = torch.randn(M, K, dtype=torch.bfloat16)
        linear = torch.nn.Linear(K, N, bias=True).to(torch.bfloat16)
        gate = torch.randn(N, dtype=torch.bfloat16)
        y = torch.randn(M, N, dtype=torch.bfloat16)

    with torch.inference_mode():

        timing = do_bench(lambda: ref_gate(x, linear, gate, y), warmup=warmups, rep=repeats) * 1e-3
        tflops = (2 * M * K * N) / 1e12 / (timing)
        print(f"PyTorch Reference: {timing * 1e3:.3f}ms, {tflops:.2f}TFLOPS")

        timing = do_bench(lambda: tk_gate(x, linear, gate, y), warmup=warmups, rep=repeats) * 1e-3
        tflops = (2 * M * K * N) / 1e12 / (timing)
        print(f"ThunderKittens: {timing * 1e3:.3f}ms, {tflops:.2f}TFLOPS")

def benchmark_gelu(M, N, K):
    with torch.device('cuda'):
        x = torch.randn(M, K, dtype=torch.bfloat16)
        linear = torch.nn.Linear(K, N, bias=True).to(torch.bfloat16)

    with torch.inference_mode():

        timing = do_bench(lambda: ref_gelu(x, linear), warmup=warmups, rep=repeats) * 1e-3
        tflops = (2 * M * K * N) / 1e12 / (timing)
        print(f"PyTorch Reference: {timing * 1e3:.3f}ms, {tflops:.2f}TFLOPS")

        timing = do_bench(lambda: tk_gelu(x, linear), warmup=warmups, rep=repeats) * 1e-3
        tflops = (2 * M * K * N) / 1e12 / (timing)
        print(f"ThunderKittens: {timing * 1e3:.3f}ms, {tflops:.2f}TFLOPS")

gelu_sizes = [
    (3072, 12288, 3072),
    (512, 12288, 3072),
    (256, 12288, 3072)
]

gated_size = [
    (3072, 3072, 12288),
    (512, 3072, 12288),
    (256, 3072, 12288),
    (3584, 3072, 15360),
    (3328, 3072, 15360)
]

print('Benchmarking Gelu')
for size in gelu_sizes:
    print('(M, N, K):', size)
    benchmark_gelu(*size)
    print()

print('Benchmarking Gated')
for size in gated_size:
    print('(M, N, K):', size)
    benchmark_gated(*size)
    print()





import sys
import os
import time
import argparse
    
from collections import defaultdict
import matplotlib.pyplot as plt
from statistics import median
import numpy as np
import math

import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange, repeat

try:
    import thunderkittens as tk
    print(f"Successfully imported thunderkittens")
except:
    print("Could not import thunderkittens")


from implementations import TaylorExp, TERMS, eps, based_kernel_test, get_based_inputs


def __eq(str, x,y, tol=1e-5, debug=False): 
    err = torch.abs(x-y).max()
    pass_str = "pass" if err < tol else "fail" 
    print(f"{str} : {pass_str} [err={err:0.5f}]")
    if(debug and (err > tol)):
        print(f"x\n{x}")
        print(f"y\n{y}")
        print(f"diff\n{x-y}")
        
    return err <= tol


def pytorch_test_v1(dt, Q, K, V, d, verbose=True, **kwargs):
    # SA: note the torch.float32 conversions are very important 

    b, h, n, D = Q.shape
    rd = math.sqrt(D) 
    rrd = math.sqrt(rd) 
    r2 = math.sqrt(2) 

    print(f"{b=}, {h=}, {n=}, {D=}")
    def make_causal(X):
        (b,h,n,m) = X.shape
        mask= ~(torch.arange(n).view(1,1,n,1) >= torch.arange(n).view(1,1,1,n)).expand(b,h,n,n)
        X[mask] = 0.
        return X

    # Overall output
    O   = torch.einsum("bhnd,bhmd->bhnm", Q.to(torch.float32), K.to(torch.float32))
    O2  = make_causal(O.to(torch.float32)**2)
    O1 = make_causal(O)
    T2  = torch.einsum("bhnm,bhmd->bhnd", O2.to(torch.float32), V.to(torch.float32))
    T2 = T2/(r2 * r2 * rd * rd)
    T1 = torch.einsum("bhnm,bhmd->bhnd", O1.to(torch.float32), V.to(torch.float32))  
    T1 = T1/rd
    T0  = V.to(torch.float32).cumsum(dim=2)

    # KV states by term (a2)
    A2 = torch.einsum("bhnd,bhnf,bhne->bhndef",K.to(torch.float32),V.to(torch.float32),K.to(torch.float32)).cumsum(dim=2) / (r2 * rd) 
    A2 = A2[:, :, -1]
    A2 = rearrange(A2, 'b h e f d -> b h (e f) d')
    K2 = torch.einsum("bhnd,bhne->bhnde", K.to(torch.float32), K.to(torch.float32)) / (rd * r2)
    Q2 = torch.einsum("bhnd,bhne->bhnde", Q.to(torch.float32), Q.to(torch.float32)) / (rd * r2)
    K2 = rearrange(K2, 'b h n d e  -> b h n ( d e )')
    Q2 = rearrange(Q2, 'b h n d e  -> b h n ( d e ) ')
    k_state_a2 = K2.to(torch.float32).cumsum(dim=2)
    D2 = torch.einsum("bhnd,bhnd->bhn", Q2.to(torch.float32), k_state_a2)

    # KV states by term (a1)
    A1 = torch.einsum("bhnd,bhne->bhnde",K.to(torch.float32),V.to(torch.float32)).cumsum(dim=2)  / rrd
    A1 = A1[:, :, -1].transpose(2, 3)
    k_state_a1 = K.to(torch.float32).cumsum(dim=2)/ (rrd)
    D1 = torch.einsum("bhnd,bhnd->bhn", Q.to(torch.float32), k_state_a1) / rrd 

    # KV states by term (a0)
    A0 = V.to(torch.float32).cumsum(dim=2)[:, :, -1]
    K0 = torch.ones(Q[..., :1].shape).to(Q.device)
    D0 =  K0.to(torch.float32).cumsum(dim=2).squeeze(-1)

    numerators   = [T0, T1, T2] 
    denominators = [D0, D1, D2]
    numerator   = sum([n for i, n in enumerate(numerators)   if i in TERMS]) 
    denominator = sum([n for i, n in enumerate(denominators) if i in TERMS]) 
    y = numerator

    kv_state = torch.cat([A0.unsqueeze(-1).transpose(2,3), A1.transpose(2,3), A2], dim=2)

    return y, kv_state


def test_correctness(): 
    dt = torch.bfloat16
    b = 2
    n = 2048
    h = 16
    d = 16
    dv = 64
    print(f"{b=}, {n=}, {d=}, {h=}")

    head_idx = torch.randint(0, h, (1,)).item()
    batch_idx = torch.randint(0, b, (1,)).item()

    Q, K, V = get_based_inputs(b, h, n, dv, dt)

    # get outputs from different methods
    pytorch_v1, kv_state_v1  = pytorch_test_v1(dt, Q, K, V, d)
    outputs, _  = based_kernel_test(dt, b, h, n, dv)
    tk_outputs, kv_state_tk = outputs[0], outputs[1]
    torch.set_printoptions(sci_mode=False)

    # check overall outputs
    print(f"Note we find numerical differences upon inspecting the tensor outputs:\n")
    print(f"Checking outputs:")
    __eq("PyTorch v1 - Based TK", pytorch_v1, tk_outputs, debug=False)

    print(pytorch_v1[1,head_idx,70:72,:4])
    print(tk_outputs[1,head_idx,70:72,:4])
    print()

    print(pytorch_v1[0,head_idx,128:130,:4])
    print(tk_outputs[0,head_idx,128:130,:4])
    print()

    print(pytorch_v1[1,head_idx,500:502,:4])
    print(tk_outputs[1,head_idx,500:502,:4])
    print()

    print("---"*10)

    # kv states
    __eq("PyTorch v1 - TK", kv_state_v1, kv_state_tk, debug=False)
    print(kv_state_v1[1,head_idx,0,:4])
    print(kv_state_tk[1,head_idx,0,:4])
    max_diff = torch.max(torch.abs(kv_state_v1 - kv_state_tk))
    pos_max_diff = int(torch.argmax(torch.abs(kv_state_v1 - kv_state_tk)))
    tk_val = kv_state_tk.cpu().float().numpy().flatten()[pos_max_diff]
    pytorch_val = kv_state_v1.cpu().float().numpy().flatten()[pos_max_diff]
    print(f"Max diff: {max_diff}; {tk_val=}, {pytorch_val=}")
    print(f"Max at position: {pos_max_diff} out of {kv_state_v1.numel()} / {kv_state_tk.numel()}")

if __name__ == "__main__":
    test_correctness()


"""
Logging utilities to make terminal slightly more delightful
"""
import rich.syntax
import rich.tree

from omegaconf import OmegaConf, DictConfig, ListConfig


def _format_arg(arg_name: str, cutoff=2) -> str:
    if arg_name is None:
        return arg_name
    arg_name = str(arg_name)
    
    # Hardcode to handle backslash
    name_splits = arg_name.split('/')
    if len(name_splits) > 1:
        return name_splits[-1]
    # Abbreviate based on underscore
    name_splits = arg_name.split('_')
    if len(name_splits) > 1:
        return ''.join([s[0] for s in name_splits])
    else:
        return arg_name[:cutoff]


def print_header(x: str) -> None:
    """
    Print a header with a line above and below
    """
    print('-' * len(x))
    print(x)
    print('-' * len(x))


def print_args(args, return_dict=False, verbose=True):
    """
    Print the arguments passed to the script
    """
    attributes = [a for a in dir(args) if a[0] != '_']
    arg_dict = {}  # switched to ewr
    if verbose:
        print('ARGPARSE ARGS')
    for ix, attr in enumerate(attributes):
        fancy = '└─' if ix == len(attributes) - 1 else '├─'
        if verbose:
            print(f'{fancy} {attr}: {getattr(args, attr)}')
        arg_dict[attr] = getattr(args, attr)
    if return_dict:
        return arg_dict


def update_description_metrics(description: str, metrics: dict):
    """
    Set the numbers that show up on progress bars
    """
    for split in metrics:
        if split != 'test':  # No look
            for metric_name, metric in metrics[split].items():
                description += f' | {split}/{metric_name}: {metric:.3f}'
    return description

        
# Control how tqdm progress bar looks        
def type_of_script():
    try:
        ipy_str = str(type(get_ipython()))
        if 'zmqshell' in ipy_str:
            return 'jupyter'
        if 'terminal' in ipy_str:
            return 'ipython'
    except:
        return 'terminal'
    
# Progress bar 
def update_pbar_display(metrics, batch_ix, pbar, prefix, batch_size, accum_iter=1):
    description = f'└── {prefix} batch {int(batch_ix)}/{len(pbar)} [batch size: {batch_size} - grad. accum. over {accum_iter} batch(es)]'
    for metric_name, metric in metrics.items():
        if metric_name == 'correct':
            description += f' | {metric_name} (acc. %): {int(metric):>5d}/{int(metrics["total"])} = {metric / metrics["total"] * 100:.3f}%'
        elif metric_name == 'acc':
            description += f' | {metric_name}: {metric:.3f}'
        elif metric_name in ['perplexity']:  # , 'bpc']:
            description += f' | {metric_name}: {Decimal(metric):.3E}'            
        elif metric_name != 'total':
            description += f' | {metric_name}: {metric / metrics["total"]:.3f}'
    pbar.set_description(description)
    
    
def print_config(config: DictConfig,
                 resolve: bool = True,
                 name: str = 'CONFIG') -> None:
    """Prints content of DictConfig using Rich library and its tree structure.
    Args:
        config (DictConfig): Configuration composed by Hydra.
        fields (Sequence[str], optional): Determines which main fields from config will
        be printed and in what order.
        resolve (bool, optional): Whether to resolve reference fields of DictConfig.
    """

    style = "bright"  # "dim"
    tree = rich.tree.Tree(name, style=style, guide_style=style)

    fields = config.keys()
    for field in fields:
        branch = tree.add(field, style=style, guide_style=style)

        config_section = config.get(field)
        branch_content = str(config_section)
        if isinstance(config_section, DictConfig):
            branch_content = OmegaConf.to_yaml(config_section, resolve=resolve)
        elif isinstance(config_section, ListConfig):
            branch_content = OmegaConf.to_yaml(config_section, resolve=resolve)

        branch.add(rich.syntax.Syntax(branch_content, "yaml"))

    rich.print(tree)


import logging
import warnings
from typing import List, Sequence

import pytorch_lightning as pl
import rich.syntax
import rich.tree
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning.utilities import rank_zero_only


# Copied from https://docs.python.org/3/howto/logging-cookbook.html#using-a-context-manager-for-selective-logging
class LoggingContext:
    def __init__(self, logger, level=None, handler=None, close=True):
        self.logger = logger
        self.level = level
        self.handler = handler
        self.close = close

    def __enter__(self):
        if self.level is not None:
            self.old_level = self.logger.level
            self.logger.setLevel(self.level)
        if self.handler:
            self.logger.addHandler(self.handler)

    def __exit__(self, et, ev, tb):
        if self.level is not None:
            self.logger.setLevel(self.old_level)
        if self.handler:
            self.logger.removeHandler(self.handler)
        if self.handler and self.close:
            self.handler.close()
        # implicit return of None => don't swallow exceptions


def get_logger(name=__name__) -> logging.Logger:
    """Initializes multi-GPU-friendly python logger."""

    logger = logging.getLogger(name)

    # this ensures all logging levels get marked with the rank zero decorator
    # otherwise logs would get multiplied for each GPU process in multi-GPU setup
    for level in ("debug", "info", "warning", "error", "exception", "fatal", "critical"):
        setattr(logger, level, rank_zero_only(getattr(logger, level)))

    return logger


def extras(config: DictConfig) -> None:
    """A couple of optional utilities, controlled by main config file:
    - disabling warnings
    - forcing debug friendly configuration
    - verifying experiment name is set when running in experiment mode
    Modifies DictConfig in place.
    Args:
        config (DictConfig): Configuration composed by Hydra.
    """

    log = get_logger(__name__)

    # disable python warnings if <config.ignore_warnings=True>
    if config.get("ignore_warnings"):
        log.info("Disabling python warnings! <config.ignore_warnings=True>")
        warnings.filterwarnings("ignore")

    # verify experiment name is set when running in experiment mode
    if config.get("experiment_mode") and not config.get("name"):
        log.info(
            "Running in experiment mode without the experiment name specified! "
            "Use `python run.py mode=exp name=experiment_name`"
        )
        log.info("Exiting...")
        exit()

    # force debugger friendly configuration if <config.trainer.fast_dev_run=True>
    # debuggers don't like GPUs and multiprocessing
    if config.trainer.get("fast_dev_run"):
        log.info("Forcing debugger friendly configuration! <config.trainer.fast_dev_run=True>")
        if config.trainer.get("gpus"):
            config.trainer.gpus = 0
        if config.datamodule.get("pin_memory"):
            config.datamodule.pin_memory = False
        if config.datamodule.get("num_workers"):
            config.datamodule.num_workers = 0


@rank_zero_only
def print_config(
    config: DictConfig,
    fields: Sequence[str] = (
        "trainer",
        "model",
        "datamodule",
        "train",
        "eval",
        "callbacks",
        "logger",
        "seed",
        "name",
    ),
    resolve: bool = True,
) -> None:
    """Prints content of DictConfig using Rich library and its tree structure.
    Args:
        config (DictConfig): Configuration composed by Hydra.
        fields (Sequence[str], optional): Determines which main fields from config will
        be printed and in what order.
        resolve (bool, optional): Whether to resolve reference fields of DictConfig.
    """

    style = "dim"
    tree = rich.tree.Tree("CONFIG", style=style, guide_style=style)

    for field in fields:
        branch = tree.add(field, style=style, guide_style=style)

        config_section = config.get(field)
        branch_content = str(config_section)
        if isinstance(config_section, DictConfig):
            branch_content = OmegaConf.to_yaml(config_section, resolve=resolve)

        branch.add(rich.syntax.Syntax(branch_content, "yaml"))

    rich.print(tree)

    with open("config_tree.txt", "w") as fp:
        rich.print(tree, file=fp)

@rank_zero_only
def print_rank_zero(*args, **kwargs) -> None:
    return print(*args, **kwargs)

def finish(
    config: DictConfig,
    model: pl.LightningModule,
    datamodule: pl.LightningDataModule,
    trainer: pl.Trainer,
    callbacks: List[pl.Callback],
    logger: List[pl.loggers.LightningLoggerBase],
) -> None:
    """Makes sure everything closed properly."""

    # without this sweeps with wandb logger might crash!
    for lg in logger:
        if isinstance(lg, pl.loggers.wandb.WandbLogger):
            import wandb
            wandb.finish()
